
/*************************************************************************/
/* MAIN.CPP                                                   20.02.2008 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#include "stdafx.h"
#include "tgd.h"

extern "C" {

tgd::DRIVER_DATA * __tavr__Engine___impl__Initialize___string___int___int(void * engine, tavr::REF_STRING name, int width, int height)
{
	tgd::g_Engine = *(__tavr_ref_t *)engine;

	HINSTANCE inst = GetModuleHandle(NULL);

	WNDCLASSEX wcls;
	ZeroMemory(&wcls, sizeof(wcls));
	wcls.cbSize			= sizeof(WNDCLASSEX);
	wcls.hInstance		= inst;
	wcls.lpszClassName	= name->value;
	wcls.style			= CS_HREDRAW | CS_VREDRAW;
	wcls.lpfnWndProc	= tgd::WndProc;
	wcls.hIcon			= LoadIcon(NULL, IDI_APPLICATION);
	wcls.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcls.hbrBackground	= (HBRUSH)GetStockObject(BLACK_BRUSH);

	if (RegisterClassEx(&wcls) == 0) {
		tgd::Log("Failed to register window class\n");
	}

	RECT rect;
	GetWindowRect(GetDesktopWindow(), &rect);
	int left = (rect.right - width) / 2;
	int top  = (rect.bottom - height) / 2;

	HWND wnd = CreateWindow(name->value, name->value, WS_OVERLAPPEDWINDOW,
		left, top, width, height, NULL, NULL, inst, NULL);

	tgd::DRIVER_DATA * res = tgd::InitializeDX(wnd, width, height);

	ShowWindow(wnd, SW_SHOW);

	return res;
}

/*************************************************************************/
/* __tavr__Engine___impl__MessageLoop()                                  */
/*                                                                       */
/*************************************************************************/
void __tavr__Engine___impl__MessageLoop___int(void * engine, tgd::DRIVER_DATA * vidData)
{
	void __tavr__Engine___impl__OnIdle(void *);

	MSG msg;
	for (;;) {
		while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
			if (msg.message == WM_QUIT) {
				return ;
			}

			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		__tavr__Engine___impl__OnIdle(engine);
		Sleep(1);
	}

	tgd::TerminateDX(vidData);
}

/*************************************************************************/
/* __tavr__Engine___impl__Release___int()                                */
/*                                                                       */
/*************************************************************************/
void __tavr__Engine___impl__Release___int(void * engine, tgd::DRIVER_DATA * vidData)
{
	tgd::TerminateDX(vidData);
}

enum {
	MOUSE_ACTION_DOWN,
	MOUSE_ACTION_UP,
	MOUSE_ACTION_MOVE,
};

enum {
	MOUSE_BUTTON_LEFT,
	MOUSE_BUTTON_RIGHT,
	MOUSE_BUTTON_MIDDLE,
};

void __tavr__Engine___impl__OnMouse___int___int___EngV2_r(void * engine, int action, int button, D3DXVECTOR2 * pos);
void __tavr__Engine___impl__OnResize___float___float___bool(void * engine, float width, float height, bool isReset);

} // extern "C"

namespace tgd {

__tavr_ref_t g_Engine;

/*************************************************************************/
/* GetMousePos()                                                         */
/*                                                                       */
/*************************************************************************/
D3DXVECTOR2 GetMousePos(HWND hwnd)
{
	RECT rect;
	GetWindowRect(hwnd, &rect);
	RECT adj = { 0, 0, 0, 0 };
	AdjustWindowRect(&adj, WS_OVERLAPPEDWINDOW, false);
	rect.left -= adj.left;
	rect.top  -= adj.top;
	rect.bottom -= adj.bottom;
	rect.right -= adj.right;

	POINT pos;
	GetCursorPos(&pos);

	pos.x -= rect.left;
	pos.y -= rect.top;
	pos.y  = rect.bottom - rect.top - pos.y;

	return D3DXVECTOR2((float)pos.x, (float)pos.y);
}

/*************************************************************************/
/* WndProc()                                                             */
/*                                                                       */
/*************************************************************************/
LRESULT WINAPI WndProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	static int curWndWidth, curWndHeight;

	switch (msg) {
		case WM_CREATE:
			{
				PostMessage(hwnd, WM_USER, 0, 0);
				RECT rect;
				GetWindowRect(hwnd, &rect);
				curWndWidth  = rect.right - rect.left;
				curWndHeight = rect.bottom - rect.top;
			}			
			break;

		case WM_USER:
		case WM_SIZE:
			{
				RECT rect;
				GetClientRect(hwnd, &rect);
				__tavr__Engine___impl__OnResize___float___float___bool(&g_Engine, (float)(rect.right - rect.left), (float)(rect.bottom - rect.top), 0);

				if (msg == WM_SIZE) {
					KillTimer(hwnd, 0);
					SetTimer(hwnd, 0, 500, NULL);
				}
			}
			break;

		case WM_TIMER:
			{
				KillTimer(hwnd, 0);

				RECT rect;
				GetWindowRect(hwnd, &rect);
				if (curWndWidth != rect.right - rect.left || curWndHeight != rect.bottom - rect.top) {
					curWndWidth  = rect.right - rect.left;
					curWndHeight = rect.bottom - rect.top;

					ResetDX(hwnd);

					GetClientRect(hwnd, &rect);
					__tavr__Engine___impl__OnResize___float___float___bool(&g_Engine, (float)(rect.right - rect.left), (float)(rect.bottom - rect.top), 1);
				}
			}
			break;

		case WM_CLOSE:
			DestroyWindow(hwnd);
			break;

		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_RBUTTONDOWN:
		case WM_RBUTTONUP:
		case WM_MBUTTONDOWN:
		case WM_MBUTTONUP:
			{
				int action = MOUSE_ACTION_UP;
				switch (msg) {
					case WM_LBUTTONDOWN:
					case WM_RBUTTONDOWN:
					case WM_MBUTTONDOWN:
						action = MOUSE_ACTION_DOWN;
						break;
				}

				int button = 0;
				switch (msg) {
					case WM_LBUTTONDOWN:
					case WM_LBUTTONUP:
						button = MOUSE_BUTTON_LEFT;
						break;
					case WM_RBUTTONDOWN:
					case WM_RBUTTONUP:
						button = MOUSE_BUTTON_RIGHT;
						break;
					case WM_MBUTTONDOWN:
					case WM_MBUTTONUP:
						button = MOUSE_BUTTON_MIDDLE;
						break;
				}

				D3DXVECTOR2 posf = GetMousePos(hwnd);
				__tavr__Engine___impl__OnMouse___int___int___EngV2_r(&g_Engine, action, button, &posf);
			}
			break;

		case WM_MOUSEMOVE:
			{
				D3DXVECTOR2 posf = GetMousePos(hwnd);
				__tavr__Engine___impl__OnMouse___int___int___EngV2_r(&g_Engine, MOUSE_ACTION_MOVE, 0, &posf);
			}
			break;

		case WM_DESTROY:
			PostQuitMessage(0);
			break;
	}

	return DefWindowProc(hwnd, msg, wparam, lparam);
}

/*************************************************************************/
/* Log()                                                                 */
/*                                                                       */
/*************************************************************************/
void Log(const char * format, ...)
{
	int size = 256;

	va_list args;
	va_start(args, format);

	for (;;) {
		char * buf = new char[size];
		if (vsnprintf(buf, size, format, args) == -1) {
			delete [] buf;
			size *= 2;
			continue;
		}
		
		OutputDebugStringA(buf);
		delete [] buf;
		break;
	}

	va_end(args);
}

/*************************************************************************/
/* Format()                                                              */
/*                                                                       */
/*************************************************************************/
std::string Format(const char * format, ...)
{
	int size = 256;

	va_list args;
	va_start(args, format);

	std::string res;

	for (;;) {
		char * buf = new char[size];
		if (vsnprintf(buf, size, format, args) == -1) {
			delete [] buf;
			size *= 2;
			continue;
		}

		res = buf;
		delete [] buf;
		break;
	}

	va_end(args);

	return res;
}

/*************************************************************************/
/* Format()                                                              */
/*                                                                       */
/*************************************************************************/
std::wstring Format(const wchar_t * format, ...)
{
	int size = 256;

	va_list args;
	va_start(args, format);

	std::wstring res;

	for (;;) {
		wchar_t * buf = new wchar_t[size];
		if (_vsnwprintf(buf, size, format, args) == -1) {
			delete [] buf;
			size *= 2;
			continue;
		}

		res = buf;
		delete [] buf;
		break;
	}

	va_end(args);

	return res;
}

} // namespace tgd

//
// End-of-file MAIN.CPP
//
