
/*************************************************************************/
/* TGD.H                                                      20.02.2008 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef TGD_H
#define TGD_H

namespace tgd {

struct DRIVER_DATA {
	IDirect3D9 *		d3d;
	IDirect3DDevice9 *	device;

	DRIVER_DATA(void)
		: d3d(), device()
	{}
};

struct TEXTURE_DATA {
	IDirect3DTexture9 * texture;
	IDirect3DDevice9 *	device;

	TEXTURE_DATA(void)
		: texture(), device()
	{}
};

struct SHADER_DATA {
	bool				isPixelShader;
	IDirect3DDevice9 *	device;
	ID3DXConstantTable *		constantTable;

	SHADER_DATA(bool isPixelShader)
		: isPixelShader(isPixelShader), constantTable(), device()
	{}
};

struct VERTEX_SHADER_DATA : SHADER_DATA {
	IDirect3DVertexShader9 *	shader;

	VERTEX_SHADER_DATA(void)
		: SHADER_DATA(false), shader()
	{}
};

struct PIXEL_SHADER_DATA : SHADER_DATA {
	IDirect3DPixelShader9 *		shader;

	PIXEL_SHADER_DATA(void)
		: SHADER_DATA(true), shader()
	{}
};

inline int FloatToByte(__tavr_float val) {
	int t = int(val * 255.0 + 0.5);
	return max(0, min(255, (int)t));
}

struct COLOR {
	float r;
	float g;
	float b;
	float a;

	DWORD toInt(void) {
		return FloatToByte(a) << 24
			|  FloatToByte(r) << 16
			|  FloatToByte(g) << 8
			|  FloatToByte(b);
	}
};

struct QUAD {
	float x;
	float y;
	float width;
	float height;
};

DRIVER_DATA *	InitializeDX		(HWND wnd, int width, int height);
void			ResetDX				(HWND wnd);
void			TerminateDX			(DRIVER_DATA * data);

void			Log					(const char * format, ...);
std::string		Format				(const char * format, ...);
std::wstring	Format				(const wchar_t * format, ...);

LRESULT WINAPI	WndProc				(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);

extern __tavr_ref_t			g_Engine;
extern IDirect3DDevice9 *	g_Device;

} // namespace tgd

#define SAFE_RELEASE(PTR) \
	if (PTR) { \
		(PTR)->Release(); \
		(PTR) = NULL; \
	}

#ifdef _DEBUG
#define DEBUG_BREAK() \
	DebugBreak()
#else
#define DEBUG_BREAK()
#endif

#define V(CMD) \
	{ \
		HRESULT _hr__ = (CMD); \
		if (FAILED(_hr__)) { \
			DEBUG_BREAK(); \
			tgd::Log("Failed: %s\n", #CMD); \
			tgd::Log("Error : %s (0x%08hd)\n", DXGetErrorStringA(_hr__), _hr__); \
			tgd::Log("Description : %s\n", DXGetErrorDescriptionA(_hr__)); \
		} \
	}

#endif // TGD_H
