
/*************************************************************************/
/* VID_DRV.CPP                                                20.02.2008 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#include "stdafx.h"
#include "tgd.h"

namespace tgd {

IDirect3DDevice9 * g_Device;

/*************************************************************************/
/* GetClosestPow2()                                                      */
/*                                                                       */
/*************************************************************************/
int GetClosestPow2(int value)
{
	int counter = -1;
	int cur = value;
	while (cur > 0) {
		cur >>= 1;
		++counter;
	}

	int res = (1 << counter);
	if (res < value) {
		return res << 1;
	}
	
	return res;
}

/*************************************************************************/
/* InitlizeDX()                                                          */
/*                                                                       */
/*************************************************************************/
DRIVER_DATA * InitializeDX(HWND hwnd, int width, int height)
{
	DRIVER_DATA * data = TAVR_NEW DRIVER_DATA();

	CoInitialize(NULL);

	data->d3d = Direct3DCreate9(D3D_SDK_VERSION);

	D3DPRESENT_PARAMETERS params;
	ZeroMemory(&params, sizeof(params));
	params.BackBufferCount			= 1;
	params.SwapEffect				= D3DSWAPEFFECT_DISCARD;
	params.hDeviceWindow			= hwnd;
	params.Windowed					= TRUE;

	data->device = NULL;
	V( data->d3d->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd, D3DCREATE_HARDWARE_VERTEXPROCESSING, &params, &data->device) );
	if (data->device == NULL) {
		MessageBoxW(hwnd, L"Failed to initialize video", L"Error", MB_OK|MB_ICONERROR);
#ifndef _DEBUG
		exit(0);
#endif
	}

	g_Device = data->device;

	return data;
}

/*************************************************************************/
/* ResetDX()                                                             */
/*                                                                       */
/*************************************************************************/
void ResetDX(HWND hwnd)
{
	D3DPRESENT_PARAMETERS params;
	ZeroMemory(&params, sizeof(params));
	params.BackBufferCount			= 1;
	params.SwapEffect				= D3DSWAPEFFECT_DISCARD;
	params.hDeviceWindow			= hwnd;
	params.Windowed					= TRUE;

	V( g_Device->Reset(&params) );
}

/*************************************************************************/
/* TerminateDX()                                                         */
/*                                                                       */
/*************************************************************************/
void TerminateDX(DRIVER_DATA * data)
{
	SAFE_RELEASE(data->device);
	SAFE_RELEASE(data->d3d);
	TAVR_DELETE(data);
}

} // namepsace tgd

extern "C" {

/*************************************************************************/
/* __tavr__Engine___impl__BeginScene()                                   */
/*                                                                       */
/*************************************************************************/
void __tavr__Engine___impl__BeginScene___int(void * engine, tgd::DRIVER_DATA * data)
{
	V( data->device->BeginScene() );
}

/*************************************************************************/
/* __tavr__Engine___impl__BeginScene()                                   */
/*                                                                       */
/*************************************************************************/
void __tavr__Engine___impl__EndScene___int(void * engine, tgd::DRIVER_DATA * data)
{
	V( data->device->EndScene() );
	V( data->device->Present(NULL, NULL, NULL, NULL) );
}

/*************************************************************************/
/* __tavr__Engine___impl__Clear___EngColor___int()                       */
/*                                                                       */
/*************************************************************************/
void __tavr__Engine___impl__Clear___int___EngColor_r(void * engine, tgd::DRIVER_DATA * data, tgd::COLOR * color)
{
	V( data->device->Clear(0, NULL, D3DCLEAR_TARGET, color->toInt(), 0, 0) );
}

/*************************************************************************/
/* __tavr__Engine___impl__EnableBlending___int___bool()                  */
/*                                                                       */
/*************************************************************************/
void __tavr__Engine___impl__EnableBlending___int___bool(void * engine, tgd::DRIVER_DATA * data, __tavr_bool enable)
{
	if (enable) {
		V( data->device->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE) );
		V( data->device->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA) );
		V( data->device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA) );
	} else {
		V( data->device->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE) );
	}
}

/*************************************************************************/
/* __tavr__Engine___impl__Clear___EngColor___int()                       */
/*                                                                       */
/*************************************************************************/
void __tavr__Engine___impl__RenderQuad___int___EngQuad_r(void * engine, tgd::DRIVER_DATA * data, tgd::QUAD * quad)
{
	V( data->device->SetFVF(D3DFVF_XYZ) );
	D3DXVECTOR3 triangles[] = {
		D3DXVECTOR3(quad->x,				quad->y					, 0),
		D3DXVECTOR3(quad->x,				quad->y + quad->height	, 0),
		D3DXVECTOR3(quad->x + quad->width,	quad->y					, 0),
		D3DXVECTOR3(quad->x + quad->width,	quad->y + quad->height	, 0),
	};
	V( data->device->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, triangles, sizeof(D3DXVECTOR3)) );
}

/*************************************************************************/
/* __tavr__Engine___impl__LoadTexture___string___int()                   */
/*                                                                       */
/*************************************************************************/
tgd::TEXTURE_DATA * __tavr__Engine___impl__LoadTexture___string___int(void * engine, tavr::REF_STRING path, tgd::DRIVER_DATA * data)
{
	tgd::TEXTURE_DATA * res = TAVR_NEW tgd::TEXTURE_DATA();
	res->device = data->device;

	res->texture = NULL;
	V( D3DXCreateTextureFromFileW(data->device, path->value, &res->texture) );
	if (res->texture == NULL) {
		std::wstring msg = tgd::Format(L"Failed to load texture '%s'", path->value);
		MessageBoxW(NULL, msg.c_str(), L"Error", MB_OK|MB_ICONERROR);
#ifndef _DEBUG
		exit(0);
#else
		tgd::Log("Failed to load texture\n");
#endif
	}

	return res;
}

/*************************************************************************/
/* __tavr__EngTexture___impl__Enable___int___int___bool()                */
/*                                                                       */
/*************************************************************************/
void __tavr__EngTexture___impl__Enable___int___int___bool(void * texture, tgd::TEXTURE_DATA * data, int source, __tavr_bool enable)
{
	if (enable) {
		V( data->device->SetTexture(source, data->texture) );
	} else {
		V( data->device->SetTexture(source, NULL) );
	}
}

/*************************************************************************/
/* __tavr__EngTexture___impl__Release___int()                            */
/*                                                                       */
/*************************************************************************/
void __tavr__EngTexture___impl__Release___int(void * tex, tgd::TEXTURE_DATA * data)
{
	SAFE_RELEASE(data->texture);

	TAVR_DELETE(data);
}

} // extern "C"

//
// End-of-file VID_DRV.CPP
//
