
/*************************************************************************/
/* VID_SHADERS.CPP                                            21.02.2008 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#include "stdafx.h"
#include "tgd.h"

extern "C" {

/*************************************************************************/
/* __tavr__Engine___impl__LoadShader___string()                          */
/*                                                                       */
/*************************************************************************/
void * __tavr__Engine___impl__LoadShader___string___int___bool(void * engine, tavr::REF_STRING path, tgd::DRIVER_DATA * driverData, __tavr_bool isPixelShader)
{
	ID3DXBuffer * code = NULL;
	ID3DXBuffer * errMsg = NULL;
	ID3DXConstantTable * constantTable = NULL;

	HRESULT hr;
	V( hr = D3DXCompileShaderFromFileW(path->value, NULL, NULL, "main",
		isPixelShader ? "ps_2_0" : "vs_2_0", 0, &code, &errMsg, &constantTable) );

	if (FAILED(hr)) {
		if (errMsg != NULL) {
			std::string msg = tgd::Format("Shader compilation failed\n\n%s", errMsg->GetBufferPointer());
			MessageBoxA(NULL, msg.c_str(), "Error", MB_OK|MB_ICONERROR);
#ifndef _DEBUG
			exit(0);
#else
			tgd::Log("Message: %s\n", errMsg->GetBufferPointer());
#endif
			SAFE_RELEASE(errMsg);
		} else {
			std::wstring msg = tgd::Format(L"Failed to load shader '%s'", path->value);
			MessageBoxW(NULL, msg.c_str(), L"Error", MB_OK|MB_ICONERROR);
#ifndef _DEBUG
			exit(0);
#else
			tgd::Log("Message: File not found");
#endif
		}
		return NULL;
	}

	if (isPixelShader) {
		tgd::PIXEL_SHADER_DATA * data = TAVR_NEW tgd::PIXEL_SHADER_DATA();
		data->constantTable = constantTable;
		data->device        = driverData->device;
		V( driverData->device->CreatePixelShader((DWORD *)code->GetBufferPointer(), &data->shader) );

		SAFE_RELEASE(code);
		return data;
	}

	tgd::VERTEX_SHADER_DATA * data = TAVR_NEW tgd::VERTEX_SHADER_DATA();
	data->constantTable = constantTable;
	data->device        = driverData->device;
	V( driverData->device->CreateVertexShader((DWORD *)code->GetBufferPointer(), &data->shader) );
	SAFE_RELEASE(code);

	return data;
}

/*************************************************************************/
/* __tavr__EngShader___impl__Enable___int___bool()                       */
/*                                                                       */
/*************************************************************************/
void __tavr__EngShader___impl__Enable___int___bool(void * engine, tgd::SHADER_DATA * data, __tavr_bool enable)
{
	if (!enable) {
		if (data->isPixelShader) {
			V( data->device->SetPixelShader(NULL) );
		} else {
			V( data->device->SetVertexShader(NULL) );
		}
		return ;
	}

	if (data->isPixelShader) {
		tgd::PIXEL_SHADER_DATA * ps = (tgd::PIXEL_SHADER_DATA *)data;
		V( data->device->SetPixelShader(ps->shader) );
	} else {
		tgd::VERTEX_SHADER_DATA * vs = (tgd::VERTEX_SHADER_DATA *)data;
		V( data->device->SetVertexShader(vs->shader) );
	}
}

/*************************************************************************/
/* __tavr__EngShader___impl__Release___int()                             */
/*                                                                       */
/*************************************************************************/
void __tavr__EngShader___impl__Release___int(void * engine, tgd::SHADER_DATA * data)
{
	SAFE_RELEASE(data->constantTable);
	// don't release data->device

	if (data->isPixelShader) {
		SAFE_RELEASE( ((tgd::PIXEL_SHADER_DATA *)data)->shader );
	} else {
		SAFE_RELEASE( ((tgd::VERTEX_SHADER_DATA *)data)->shader );
	}


	TAVR_DELETE(data);
}

/*************************************************************************/
/* __tavr__EngShader___impl__Release___int()                             */
/*                                                                       */
/*************************************************************************/
void __tavr__EngShader___impl__SetConstant___int___string___float___float___float___float(void * engine, tgd::SHADER_DATA * data, tavr::REF_STRING name, float a, float b, float c, float d)
{
	char * buf = (char *)alloca(name->length + 1);
	wcstombs(buf, name->value, name->length);
	buf[name->length] = 0;

	D3DXHANDLE handle = data->constantTable->GetConstantByName(NULL, buf);

	if (handle == NULL) {
		tgd::Log("Unknown pixel shader constant '%s'\n", buf);
		DEBUG_BREAK();

		std::string msg = tgd::Format("Unknown pixel shader constant '%s'", buf);
		MessageBoxA(NULL, msg.c_str(), "Error", MB_OK|MB_ICONERROR);
#ifndef _DEBUG
		exit(0);
#endif
		return ;
	}

	float vals[] = { a, b, c, d };
	V( data->constantTable->SetFloatArray(data->device, handle, vals, 4) );
}

/*************************************************************************/
/* __tavr__EngShader___impl__Release___int()                             */
/*                                                                       */
/*************************************************************************/
void __tavr__EngShader___impl__SetConstant___int___string___float___float___float(void * engine, tgd::SHADER_DATA * data, tavr::REF_STRING name, float a, float b, float c)
{
	char * buf = (char *)alloca(name->length + 1);
	wcstombs(buf, name->value, name->length);
	buf[name->length] = 0;

	D3DXHANDLE handle = data->constantTable->GetConstantByName(NULL, buf);

	if (handle == NULL) {
		tgd::Log("Unknown pixel shader constant '%s'\n", buf);
		DEBUG_BREAK();

		std::string msg = tgd::Format("Unknown pixel shader constant '%s'", buf);
		MessageBoxA(NULL, msg.c_str(), "Error", MB_OK|MB_ICONERROR);
#ifndef _DEBUG
		exit(0);
#endif
		return ;
	}

	float vals[] = { a, b, c };
	V( data->constantTable->SetFloatArray(data->device, handle, vals, 3) );
}

/*************************************************************************/
/* __tavr__EngShader___impl__Release___int()                             */
/*                                                                       */
/*************************************************************************/
void __tavr__EngShader___impl__SetConstant___int___string___float___float(void * engine, tgd::SHADER_DATA * data, tavr::REF_STRING name, float a, float b)
{
	char * buf = (char *)alloca(name->length + 1);
	wcstombs(buf, name->value, name->length);
	buf[name->length] = 0;

	D3DXHANDLE handle = data->constantTable->GetConstantByName(NULL, buf);

	if (handle == NULL) {
		tgd::Log("Unknown pixel shader constant '%s'\n", buf);
		DEBUG_BREAK();

		std::string msg = tgd::Format("Unknown pixel shader constant '%s'", buf);
		MessageBoxA(NULL, msg.c_str(), "Error", MB_OK|MB_ICONERROR);
#ifndef _DEBUG
		exit(0);
#endif
		return ;
	}

	float vals[] = { a, b };
	V( data->constantTable->SetFloatArray(data->device, handle, vals, 2) );
}

/*************************************************************************/
/* __tavr__EngShader___impl__Release___int()                             */
/*                                                                       */
/*************************************************************************/
void __tavr__EngShader___impl__SetConstant___int___string___float(void * engine, tgd::SHADER_DATA * data, tavr::REF_STRING name, float a)
{
	char * buf = (char *)alloca(name->length + 1);
	wcstombs(buf, name->value, name->length);
	buf[name->length] = 0;

	D3DXHANDLE handle = data->constantTable->GetConstantByName(NULL, buf);

	if (handle == NULL) {
		tgd::Log("Unknown pixel shader constant '%s'\n", buf);
		DEBUG_BREAK();

		std::string msg = tgd::Format("Unknown pixel shader constant '%s'", buf);
		MessageBoxA(NULL, msg.c_str(), "Error", MB_OK|MB_ICONERROR);
#ifndef _DEBUG
		exit(0);
#endif
		return ;
	}

	V( data->constantTable->SetFloat(data->device, handle, a) );
}

} // extern "C"

//
// End-of-file VID_SHADERS.CPP
//
