/************************************************************
 * TAVR.HPP                                        16.10.2007
 *
 * .
 ************************************************************/

/************************************************************************/
/* tavr::TYPEINFO::IsEqual
/*
/************************************************************************/
bool TYPEINFO::IsEqual(const TYPEINFO & p) const
{
	if (type != p.type) {
		return false;
	}

	if (isRef != p.isRef || array != p.array) {
		return false;
	}

	switch (type) {
		case TYPEINFO::TYPE_TYPE:
			return info.type == p.info.type;

		default:
			if (IsValue()) {
				return info.type == p.info.type;
			}
			assert(!"Unknown TYPEINFO type");
			break;
	}

	return false;
}

/************************************************************************/
/* tavr::TYPEINFO::IsTypeEqual
/*
/************************************************************************/
bool TYPEINFO::IsTypeEqual(const TYPEINFO & p) const
{
	assert(type   == TYPEINFO::TYPE_TYPE || IsValue());
	assert(p.type == TYPEINFO::TYPE_TYPE || p.IsValue());

	if (info.type->subtype   == tavr::TYPE::SUBTYPE_NULL && p.IsTypeByRef()
	||  p.info.type->subtype == tavr::TYPE::SUBTYPE_NULL && IsTypeByRef()) {
		return true;
	}

	// ignore 'ref' property
	return array == p.array && info.type == p.info.type;
}

/************************************************************************/
/* tavr::TYPEINFO::SetType
/*
/************************************************************************/
void TYPEINFO::SetType(tavr::TYPE * type_)
{
	type		= TYPEINFO::TYPE_TYPE;
	info.type	= type_;
	array		= 0;
	isRef		= false;
}

/************************************************************************/
/* tavr::TYPEINFO::IsUserType
/*
/************************************************************************/
bool TYPEINFO::IsUserType(void) const
{
	if (!IsValue() && type != TYPEINFO::TYPE_TYPE) {
		return false;
	}

	if (array != 0) {
		return false;
	}

	switch (info.type->subtype) {
		case tavr::TYPE::SUBTYPE_STRUCT:
		case tavr::TYPE::SUBTYPE_AUTO:
		case tavr::TYPE::SUBTYPE_EVENT:
			return true;
	}
	return false;
}

/************************************************************************/
/* tavr::TYPEINFO::IsBoolean
/*
/************************************************************************/
bool TYPEINFO::IsBoolean(void) const
{
	if (!IsValue() && type != TYPEINFO::TYPE_TYPE) {
		return false;
	}

	return array == 0 && info.type->subtype == tavr::TYPE::SUBTYPE_BOOLEAN;
}

/************************************************************************/
/* tavr::TYPEINFO::IsCharacter
/*
/************************************************************************/
bool TYPEINFO::IsCharacter(void) const
{
	if (!IsValue() && type != TYPEINFO::TYPE_TYPE) {
		return false;
	}

	return array == 0 && info.type->subtype == tavr::TYPE::SUBTYPE_CHARACTER;
}

/************************************************************************/
/* tavr::TYPEINFO::IsIntegral
/*
/************************************************************************/
bool TYPEINFO::IsIntegral(void) const
{
	if (!IsValue() && type != TYPEINFO::TYPE_TYPE) {
		return false;
	}

	if (array != 0) {
		return false;
	}

	return info.type->subtype == tavr::TYPE::SUBTYPE_INTEGRAL;
}

/************************************************************************/
/* tavr::TYPEINFO::IsFloating
/*
/************************************************************************/
bool TYPEINFO::IsFloating(void) const
{
	if (!IsValue() && type != TYPEINFO::TYPE_TYPE) {
		return false;
	}

	if (array != 0) {
		return false;
	}

	return info.type->subtype == tavr::TYPE::SUBTYPE_FLOATING;
}

/*************************************************************************/
/* TYPEINFO::IsStruct()                                                  */
/*                                                                       */
/*************************************************************************/
bool TYPEINFO::IsStruct(void) const
{
	return !IsArray()
		&& (type == TYPE_TYPE || IsValue())
		&& info.type->subtype == tavr::TYPE::SUBTYPE_STRUCT;
}

/*************************************************************************/
/* TYPEINFO::IsAuto()                                                    */
/*                                                                       */
/*************************************************************************/
bool TYPEINFO::IsAuto(void) const
{
	return !IsArray()
		&& (type == TYPE_TYPE || IsValue())
		&& info.type->subtype == tavr::TYPE::SUBTYPE_AUTO;
}

/*************************************************************************/
/* TYPEINFO::IsEvent()                                                   */
/*                                                                       */
/*************************************************************************/
bool TYPEINFO::IsEvent(void) const
{
	return !IsArray() 
		&& (type == TYPE_TYPE || IsValue())
		&& info.type->subtype == tavr::TYPE::SUBTYPE_EVENT; 
}


/************************************************************************/
/* tavr::TYPEINFO::SetRValue
/*
/************************************************************************/
void TYPEINFO::SetRValue(tavr::TYPE * type_)
{
	type		= TYPEINFO::TYPE_R_VALUE;
	info.type	= type_;
	array		= 0;
	isRef		= false;
}

/************************************************************************/
/* tavr::TYPEINFO::SetLValue
/*
/************************************************************************/
void TYPEINFO::SetLValue(tavr::TYPE * type_)
{
	type		= TYPEINFO::TYPE_L_VALUE;
	info.type	= type_;
	array		= 0;
	isRef		= false;
}

/************************************************************************/
/* tavr::TYPEINFO::Clear
/*
/************************************************************************/
void TYPEINFO::Clear(void)
{
	memset(this, 0, sizeof(TYPEINFO));
}

/************************************************************************/
/* tavr::TYPE::IsTypeByValue
/*
/************************************************************************/
bool TYPE::IsTypeByValue(void) const
{
	return subtype == tavr::TYPE::SUBTYPE_STRUCT;
}

/************************************************************************/
/* tavr::TYPEINFO::IsTypeByValue
/*
/************************************************************************/
bool TYPEINFO::IsTypeByValue(void) const
{
	if (!IsValue() && type != TYPEINFO::TYPE_TYPE) {
		return false;
	}

	if (array != 0) {
		return false;
	}

	return info.type->IsTypeByValue();
}

/************************************************************************/
/* tavr::TYPE::IsTypeByRef
/*
/************************************************************************/
bool TYPE::IsTypeByRef(void) const
{
	switch (subtype) {
		case tavr::TYPE::SUBTYPE_AUTO:
		case tavr::TYPE::SUBTYPE_EVENT:
		case tavr::TYPE::SUBTYPE_NULL:
			return true;
	}
	return false;
}

/************************************************************************/
/* tavr::TYPEINFO::IsTypeByRef
/*
/************************************************************************/
bool TYPEINFO::IsTypeByRef(void) const
{
	if (!IsValue() && type != TYPEINFO::TYPE_TYPE) {
		return false;
	}

	if (array != 0) {
		return true; // arrays are types by ref too
	}

	return info.type->IsTypeByRef();
}

/************************************************************************/
/* tavr::TYPEINFO::IsNull
/*
/************************************************************************/
bool TYPEINFO::IsNull(void) const
{
	 return (type == tavr::TYPEINFO::TYPE_TYPE || IsValue())
		 && info.type->subtype == tavr::TYPE::SUBTYPE_NULL;
}

/************************************************************************/
/* tavr::METHOD::operator==
/*
/************************************************************************/
bool METHOD::operator == (const METHOD & p) const
{
	if (name != p.name) {
		return false;
	}

	if (args.size() != p.args.size()) {
		return false;
	}

	for (int i=0; i<(int)args.size(); ++i) {
		if (!args[i].type.IsEqual(p.args[i].type)) {
			return false;
		}
	}

	return true;
}

/************************************************************************/
/* tavr::TYPEINFO::operator<
/*
/************************************************************************/
bool TYPEINFO::operator < (const TYPEINFO & ti) const
{
	if (type != ti.type) {
		return type < ti.type;
	}
	if (array != ti.array) {
		return array < ti.array;
	}
	if (isRef != ti.isRef) {
		return isRef < ti.isRef;
	}

	// yep, that's a hack
	return info.type < ti.info.type;
}

/************************************************************************/
/* tavr::LOOKUP_CONTEXT::GetParentNamespace
/*
/************************************************************************/
NAMESPACE * LOOKUP_CONTEXT::GetParentNamespace(void) const
{
	if (curNamespace != NULL) {
		return curNamespace;
	}

	TYPE * type = NULL;

	if (curType != NULL) {
		type = curType;
	} else if (curMethod != NULL) {
		if (curMethod->IsProcedure()) {
			return curMethod->owner.names;
		}
		assert(curMethod->IsMethod());

		type = curMethod->owner.type;
	}

	// TODO: support subtypes
	if (type != NULL) {
		assert(type->parentNamespace != NULL);
		return type->parentNamespace;
	}

	return NULL;
}

//
// End-of-file TAVR.HPP
//
