/************************************************************
 * TAVR_PROJECT_LOADER.H                           09.10.2007
 *
 * .
 ************************************************************/
#ifndef _TAVR_PROJECT_LOADER_H_
#define _TAVR_PROJECT_LOADER_H_

#include "tavr.h"
#include "tavr_project_loader_impl.h"
#include "tavr_project_loader_lex.h"

namespace ecg {

class TAVR_PROJECT_LOADER : public ecg::HELPER<TAVR_PROJECT_LOADER, TAVR_PROJECT_LOADER_IMPL, TAVR_PROJECT_LOADER_LEX> {
public:
	TAVR_PROJECT_LOADER(tavr::PROJECT * project)
		: project(project)
	{}

public:
	inline void					BP							(void) // begin project
	{
		assert(items.empty());
		tavr::PROJECT_SETTINGS * item = project->alloc.Alloc<tavr::PROJECT_SETTINGS> ();
		items.push(item);
	}

	inline void					Project					(void)
	{
		assert(items.size() == 1);
		project->settings = items.top();
		items.pop();
	}

	inline void					AddSection				(const std::string & name)
	{
		tavr::PROJECT_SETTINGS * item = items.top();
		items.pop();
		items.top()->subSettings.push_back(item);
	}

	inline void					BS							(const std::string & name) // begin section
	{
		tavr::PROJECT_SETTINGS * item = project->alloc.Alloc<tavr::PROJECT_SETTINGS>();
		item->name = name;
		items.push(item);
	}

	inline void					AddValue					(const std::string & left, const std::string & right)
	{
		tavr::PROJECT_SETTINGS * item = project->alloc.Alloc<tavr::PROJECT_SETTINGS>();
		item->name			= left;
		item->value			= right;
		items.top()->subSettings.push_back(item);
	}

	inline void					ProcessLiteral			(const std::string & value, std::string & res)
	{
		assert(value.size() >= 2);

		res.reserve(value.size() - 2);

		bool slash = false;
		for (int i=1; i<(int)value.size() - 1; ++i) {
			if (!slash) {
				if (value[i] == '\\') {
					slash = true;
					continue;
				}
				res += value[i];
				continue;
			}

			res += value[i];
			slash = false;
		}
	}

private:
	tavr::PROJECT *			project;
	std::stack<tavr::PROJECT_SETTINGS *> items;
};

} // namespace ecg

#endif // _TAVR_PROJECT_LOADER_H_
