
/*************************************************************************/
/* TAVR_PROJECT_LOADER_IMPL.H                                        E13 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef TAVR_PROJECT_LOADER_IMPL_H
#define TAVR_PROJECT_LOADER_IMPL_H

#ifndef TOKEN_EOF
#define TOKEN_EOF		-1
#endif

#ifndef TOKEN_ERROR
#define TOKEN_ERROR		-2
#endif

namespace ecg {

class TAVR_PROJECT_LOADER;

/*************************************************************************/
/* class TAVR_PROJECT_LOADER_IMPL                                        */
/*                                                                       */
/*************************************************************************/
class TAVR_PROJECT_LOADER_IMPL {
public:
	enum {
		STATE_COUNT		= 15,
		TERM_COUNT		= 9,
		NONTERM_COUNT	= 7,
		RULE_COUNT		= 10,
	};

	//enum TERMINALS {
	//		TOKEN_T_NAME_1,				// "{"
	//		TOKEN_T_NAME_2,				// "}"
	//		TOKEN_T_NAME_3,				// "="
	//		TOKEN_T_NAME_4,				// ";"
	//		TOKEN_T_STRING,				// $string
	//		TOKEN_T_LITERAL,			// $literal
	//		TOKEN_T_CLINE,				// $cline
	//		TOKEN_T_CBEGIN,				// $cbegin
	//		TOKEN_T_CEND,				// $cend
	//};

	enum NONTERMINALS {
		TOKEN_NT_ANONYMOUS_0,	// anonymous_0
		TOKEN_NT_DECL_LIST,		// decl_list
		TOKEN_NT_PROJECT,		// project
		TOKEN_NT_DECL,			// decl
		TOKEN_NT_WORD,			// word
		TOKEN_NT_ANONYMOUS_1,	// anonymous_1
		TOKEN_NT_PROJECT_1,		// project_1
	};

	typedef int VALUE_T;
	typedef TAVR_PROJECT_LOADER IMPL_T;

public:
						TAVR_PROJECT_LOADER_IMPL (void);
	void				Compile				(void);

protected:
	void				intPushString		(void);
	void				intPushEmptyString	(void);
	void				intCreateStringArray(void);
	void				intAppendStringArray(void);

	template <class T> void intCreatePtrArray(void);
	template <class T> void intAppendPtrArray(void);
	template <class T> void intCreateValArray(void);
	template <class T> void intAppendValArray(void);

protected:
	void				AddSection_Impl0		(void);
	void				AddValue_Impl0		(void);
	void				BS_Impl0		(void);
	void				ProcessLiteral_Impl0		(void);

protected:
	COMPILER_STATE_F	compilerState;
	PTR_STACK			pstack;
	std::stack<std::string> sstack;
	RAW_STACK			vstack;

private:
	IMPL_T &			Impl			(void);

private:
	typedef AUTO_CONFIG<STATE_COUNT, TERM_COUNT, NONTERM_COUNT, RULE_COUNT> CONFIG_T;
	typedef AUTO<IMPL_T, CONFIG_T> AUTO_T;
	friend AUTO_T;

	typedef AUTO_T::ACTION		ACTION;
	typedef AUTO_T::ACTION_PROC	ACTION_PROC;
	typedef AUTO_T::PRODUCTION	PRODUCTION;

	static const ACTION			actionTable			[STATE_COUNT][TERM_COUNT];
	static const ACTION			actionTableEof		[STATE_COUNT];
	static const ACTION			actionTableError	[STATE_COUNT];
	static const int			productionTable		[STATE_COUNT][NONTERM_COUNT];
	static const PRODUCTION		procTableProduction	[RULE_COUNT];
	static const ACTION_PROC	procTableTerm		[TERM_COUNT];
	static const ACTION_PROC	procTableNonterm	[NONTERM_COUNT];

	AUTO_T				compilerAuto;
};

} // namespace ecg

#endif // TAVR_PROJECT_LOADER_IMPL_H

