/************************************************************
 * TAVR_RUNTIME.CPP                                10.10.2007
 *
 * .
 ************************************************************/

#include "tavr_runtime.hpp"
#include <windows.h>

using namespace tavr;

//#define PRINT_ALLOCATIONS
//#define PRINT_STACK_TRACE

#ifdef PRINT_STACK_TRACE
#include "win32_stack_trace.h"
#endif

#ifdef __TAVR_ENABLE_MEM_STAT
STAT_INFO_SET tavr::memStatInfo;
#endif

//////////////////////////////////////////////////////////////////////////
// Event procs
__tavr_ref_t __tavr_cur_event;

static int pendingQueueIndex = -1;
static std::vector<PENDING_QUEUE> pendingQueues;

/*************************************************************************/
/* __tavr_post_event()                                                   */
/*                                                                       */
/*************************************************************************/
void __tavr_post_event(__tavr_ref_t event, __tavr_ref_t automaton, __tavr_auto_proc_t autoProc)
{
	__TAVR_ASSERT(pendingQueueIndex >= 0, "Internall error");
	
	PENDING_EVENT pe = { event, automaton, autoProc };
	pendingQueues[pendingQueueIndex].push(pe);
}

/*************************************************************************/
/* __tavr_process_event_queue()                                          */
/*                                                                       */
/*************************************************************************/
void __tavr_process_event_queue(PENDING_QUEUE & queue)
{
	while (!queue.empty()) {
		PENDING_EVENT & p = queue.front();
		__tavr_cur_event = p.event;
		p.proc(&p.dstAuto);
		TAVR_DEC_REF(__tavr_cur_event);
		__tavr_cur_event.ptr = __tavr_cur_event.ref = NULL;
		TAVR_DEC_REF(p.dstAuto);
		queue.pop();
	}
}

/*************************************************************************/
/* __tavr_add_post_layer()                                               */
/*                                                                       */
/*************************************************************************/
void __tavr_add_post_layer(void)
{
	++pendingQueueIndex;
	if (pendingQueueIndex >= (int)pendingQueues.size()) {
		pendingQueues.push_back(PENDING_QUEUE());
	}
}

/*************************************************************************/
/* __tavr_remove_post_layer()                                            */
/*                                                                       */
/*************************************************************************/
void __tavr_remove_post_layer(void)
{
	__tavr_process_event_queue(pendingQueues[pendingQueueIndex]);
	--pendingQueueIndex;
}

/*************************************************************************/
/* __tavr_out_call()                                                     */
/*                                                                       */
/*************************************************************************/
void __tavr_out_call(void * subscribeList, __tavr_ref_t event)
{
	SUBSCRIBED_AUTO * cur = (SUBSCRIBED_AUTO *)subscribeList;
	
	while (cur != NULL) {
		if (cur->isSend) {
			__tavr_cur_event = event;
			TAVR_INC_REF(__tavr_cur_event);

			++cur->used;
			TAVR_INC_REF(cur->srcAuto);
			cur->proc(&cur->srcAuto);
			TAVR_DEC_REF(cur->srcAuto);
			--cur->used;

			TAVR_DEC_REF(__tavr_cur_event);
			__tavr_cur_event.ptr = __tavr_cur_event.ref = NULL;
		} else {
			TAVR_INC_REF(event);
			TAVR_INC_REF(cur->srcAuto);
			__tavr_post_event(event, cur->srcAuto, cur->proc);
		}

		SUBSCRIBED_AUTO * tmp = cur;
		cur = cur->next;
		if (tmp->used < 0) {
			TAVR_DELETE(tmp);
		}
	}

	TAVR_DEC_REF(event);
}

/*************************************************************************/
/* __tavr_add_out_list()                                                 */
/*                                                                       */
/*************************************************************************/
void __tavr_add_out_list(void ** subscribeList, __tavr_ref_t automaton, __tavr_auto_proc_t autoProc, int isPost)
{
	SUBSCRIBED_AUTO * sa = TAVR_NEW SUBSCRIBED_AUTO;
	sa->srcAuto		= automaton;
	sa->proc		= autoProc;
	sa->isSend		= !isPost;
	sa->used		= 0;
	sa->next		= NULL;

	SUBSCRIBED_AUTO ** prev = (SUBSCRIBED_AUTO **)subscribeList;
	SUBSCRIBED_AUTO * cur   = *prev;
	while (cur != NULL && cur->srcAuto.ref != automaton.ref) {
		prev = &cur->next;
		cur = cur->next;
	}

	if (cur == NULL) {
		*prev = sa;
	} else {
		TAVR_DEC_REF(automaton);
		TAVR_DELETE(sa);
	}
}

/*************************************************************************/
/* __tavr_remove_out_list()                                              */
/*                                                                       */
/*************************************************************************/
void __tavr_remove_out_list(void ** subscribeList, __tavr_ref_t automaton)
{
	SUBSCRIBED_AUTO ** prev = (SUBSCRIBED_AUTO **)subscribeList;
	SUBSCRIBED_AUTO * cur   = *prev;
	while (cur != NULL && cur->srcAuto.ref != automaton.ref) {
		prev = &cur->next;
		cur = cur->next;
	}

	if (cur != NULL) {
		*prev = cur->next;
		
		if (--cur->used < 0) {
			TAVR_DELETE(cur);
		}

		TAVR_DEC_REF(automaton);
	}
	TAVR_DEC_REF(automaton);
}

/*************************************************************************/
/* __tavr_release_out_list()                                             */
/*                                                                       */
/*************************************************************************/
void __tavr_release_out_list(void ** subscribeList)
{
	SUBSCRIBED_AUTO * cur   = *(SUBSCRIBED_AUTO **)subscribeList;
	*subscribeList = NULL;

	while (cur != NULL) {
		SUBSCRIBED_AUTO * tmp = cur;
		cur = cur->next;

		TAVR_DEC_REF(tmp->srcAuto);
		TAVR_DELETE(tmp);
	}
}

/*************************************************************************/
/* __tavr_tirt_destr()                                                   */
/*                                                                       */
/*************************************************************************/
void __tavr_tirt_destr(__tavr_ref_t * p)
{
	TAVR_DEC_REF(*p);
}
void * __tavr_tirt_ref_array[] = {
	"ref[]",
	__tavr_tirt_destr,
};

/************************************************************************/
/* main
/*
/************************************************************************/
int main(int argc, const char * argv[])
{
	srand(timeGetTime());

	int mainRes = 0;
	{
		// prepare args 
		__tavr_int * arrData = (__tavr_int *)__tavr_alloc_array(argc - 1, sizeof(__tavr_ref_t), __tavr_tirt_ref_array __TAVR_MEM_STAT_FILE_LINE);
		memset(arrData + 2, 0, (argc - 1) * sizeof(__tavr_ref_t));
		REF_ARRAY<REF_STRING> arr = arrData;

		for (int i=1; i<argc; ++i) {
			REF_STRING arg = __tavr_alloc_char_string((char *)argv[i] __TAVR_MEM_STAT_FILE_LINE);
			arr[i - 1] = arg;
		}

		int __tavr__main___string_a(__tavr_ref_t);
		mainRes = __tavr__main___string_a(arr);
	}

#ifdef __TAVR_ENABLE_MEM_STAT
	if (memStatInfo.empty()) {
		return mainRes;
	}

	printf("\n--------------------------\nMemory leaks found!\n");
	int total = 0;
	for (STAT_INFO_SET::const_iterator iter = memStatInfo.begin(); iter != memStatInfo.end(); ++iter) {
		printf("0x%08x size=%d refs=%d\t%s(%d)\n", (int)iter->ptr, iter->size, *(int*)iter->ptr, iter->file, iter->line);
		total += iter->size;
	}
	printf("Total: %d\n", total);
#endif

	return mainRes;
}

/*************************************************************************/
/* WinMain()                                                             */
/*                                                                       */
/*************************************************************************/
int WINAPI WinMain(__in HINSTANCE hInstance, __in_opt HINSTANCE hPrevInstance, __in_opt LPSTR lpCmdLine, __in int nShowCmd )
{
	// TODO: support arguments
	const char * args[] = { "" };
	return main(1, args);
}

/**************************************************************************
* __tavr_alloc()                                                          *
*                                                                         *
**************************************************************************/
void * __tavr_alloc(__tavr_int size, void * typeinfo __TAVR_MEM_STAT_INFO)
{
	ECG_STATIC_ASSERT(sizeof(void *) == sizeof(__tavr_int));
	__tavr_int * res = (__tavr_int *)Alloc(size + 2 * sizeof(__tavr_int) __TAVR_MEM_STAT_VALUE);
	res[0] = 1;
	res[1] = (__tavr_int)typeinfo;
	return res;
}

/**************************************************************************
* __tavr_free()                                                           *
*                                                                         *
**************************************************************************/
void __tavr_free(void * ptr __TAVR_MEM_STAT_INFO)
{
	Free(ptr __TAVR_MEM_STAT_VALUE);
}

/**************************************************************************
* __tavr_alloc_array()                                                    *
*                                                                         *
**************************************************************************/
void * __tavr_alloc_array(__tavr_int count, __tavr_int size, void * typeinfo __TAVR_MEM_STAT_INFO)
{
	__tavr_int * res = (__tavr_int *)Alloc(count * size + __TAVR_ARRAY_OFFSET * sizeof(__tavr_int) __TAVR_MEM_STAT_VALUE);
	res[0] = 1;
	res[1] = (__tavr_int)typeinfo;
	res[2] = count;
	return res;
}

/**************************************************************************
* __tavr_free_array()                                                     *
*                                                                         *
**************************************************************************/
void __tavr_free_array(void * ptr __TAVR_MEM_STAT_INFO)
{
	Free(ptr __TAVR_MEM_STAT_VALUE);
}

/**************************************************************************
* __tavr_assert()                                                         *
*                                                                         *
**************************************************************************/
void __tavr_assert(char * condition, char * filename, int line)
{
	printf("\nAssertion failed!\nInfo: %s\nFile: %s\nLine: %d\n", condition, filename, line);
#ifdef _DEBUG
	DebugBreak();
#endif
#ifdef PRINT_STACK_TRACE
	StackTraceEngine::PrintStackTrace();
#endif
	exit(0);
}

//
// End-of-file TAVR_RUNTIME.CPP
//
