/************************************************************
 * TAVR_RUNTIME.H                                  08.10.2007
 *
 * .
 ************************************************************/
#ifndef _TAVR_RUNTIME_H_
#define _TAVR_RUNTIME_H_

#include <memory.h>

#ifdef __cplusplus
extern "C" {
#endif

#pragma warning(disable : 4312)
#pragma pack(1)

#define __TAVR_ENABLE_MEM_STAT

typedef unsigned char		__tavr_bool;
typedef unsigned short		__tavr_char;

typedef unsigned char		__tavr_byte;
typedef signed char			__tavr_sbyte;
typedef unsigned short		__tavr_word;
typedef signed short		__tavr_sword;
typedef int					__tavr_int;		// 32 bit integer
typedef unsigned int		__tavr_uint;	
typedef __int64				__tavr_long;	// 64 bit integer
typedef unsigned __int64	__tavr_ulong;

typedef float				__tavr_float;	// 32 bit float
typedef double				__tavr_double;	// 64 bit float

typedef struct {
	__tavr_int *	ref;
	void *			ptr;
} __tavr_ref_t;

#define __TAVR_RELAX_CASE_DEFAULT() \
	case 0: __tavr_this->__tavr_relax |= __TAVR_RELAX_BUSY_BIT; break;

#define __TAVR_RELAX_CASE(INDEX, LABEL) \
	case INDEX: __tavr_this->__tavr_relax |= __TAVR_RELAX_BUSY_BIT;  goto LABEL;

#define __TAVR_RELAX_CASE_DEFAULT2() \
	case 0: __tavr_this->ptr->__tavr_relax |= __TAVR_RELAX_BUSY_BIT; break;

#define __TAVR_RELAX_CASE2(INDEX, LABEL) \
	case INDEX: __tavr_this->ptr->__tavr_relax |= __TAVR_RELAX_BUSY_BIT;  goto LABEL;

#define __TAVR_RELAX_CASE_DEFAULT_S(ID) \
	case 0: __tavr_this->__tavr_relax_##ID |= __TAVR_RELAX_BUSY_BIT; break;

#define __TAVR_RELAX_CASE_S(ID, INDEX, LABEL) \
	case INDEX: __tavr_this->__tavr_relax_##ID |= __TAVR_RELAX_BUSY_BIT;  goto LABEL;

#define __TAVR_RELAX_CASE_DEFAULT2_S(ID) \
	case 0: __tavr_this->ptr->__tavr_relax_##ID |= __TAVR_RELAX_BUSY_BIT; break;

#define __TAVR_RELAX_CASE2_S(ID, INDEX, LABEL) \
	case INDEX: __tavr_this->ptr->__tavr_relax_##ID |= __TAVR_RELAX_BUSY_BIT;  goto LABEL;

#define __TAVR_RELAX_BUSY_BIT 0x80000000

//////////////////////////////////////////////////////////////////////////
// Ref Inc/Dec

typedef void (*__tavr_destr_t)(void * p);

#define __TAVR_INC_REF(REF) \
	if ((REF).ref) { \
		/*printf("inc_ref(%d) 0x%08x (%s, %d)\n", (REF).ref[0], (int)(REF).ref, __FILE__, __LINE__);*/ \
		++*(REF).ref; \
	}

#define __TAVR_DEC_REF(REF) \
	if ((REF).ref) { \
		/*printf("dec_ref(%d) 0x%08x (%s, %d)\n", (REF).ref[0], (int)(REF).ref, __FILE__, __LINE__);*/ \
		if (!--(*(REF).ref)) { \
			((__tavr_destr_t)((void **)(REF).ref[1])[1])(&REF); \
			__tavr_free((REF).ref); \
		} \
	}

//////////////////////////////////////////////////////////////////////////
// Array procs

#define __TAVR_ARRAY_LENGTH(REF) \
	(REF).ref[2]

#define __TAVR_DESTR_ARRAY_REF(REF) \
	while (__TAVR_ARRAY_LENGTH(REF) > 0) { \
		-- __TAVR_ARRAY_LENGTH(REF); \
		__TAVR_DEC_REF(REF.ptr[__TAVR_ARRAY_LENGTH(REF)]); \
	}

#define __TAVR_DESTR_ARRAY_VALUE(REF, DESTR) \
	while (__TAVR_ARRAY_LENGTH(REF) > 0) { \
		-- __TAVR_ARRAY_LENGTH(REF); \
		DESTR(&REF.ptr[__TAVR_ARRAY_LENGTH(REF)]); \
	}

///////////////////////////////////////////////////////////////////////////
// Asserts

#define __TAVR_ASSERT(COND, INFO) \
	if (!(COND)) { \
		__tavr_assert(INFO, __FILE__, __LINE__); \
	} 

#define __TAVR_ASSERT_NULL_PTR(REF) \
	__TAVR_ASSERT((REF).ptr != 0, "Dereferencing null pointer")

#define __TAVR_ASSERT_ARRAY_INDEX(REF, INDEX) \
	__TAVR_ASSERT_NULL_PTR(REF); \
	__TAVR_ASSERT((INDEX) >= 0 && (INDEX) < __TAVR_ARRAY_LENGTH(REF), "Array out of bounds");

#define __TAVR_ASSERT_STRING_INDEX(REF, INDEX) \
	__TAVR_ASSERT_NULL_PTR(REF); \
	__TAVR_ASSERT((INDEX) >= 0 && (INDEX) < (REF).ptr->length, "String out of bounds");

#define __TAVR_UNKNOWN_RELAX_CONTEXT() \
	__TAVR_ASSERT(0, "Invalid relax context");

//////////////////////////////////////////////////////////////////////////
// Event and send/post data
extern	__tavr_ref_t __tavr_cur_event;
typedef void (* __tavr_auto_proc_t)	(__tavr_ref_t *);

void		__tavr_post_event			(__tavr_ref_t event, __tavr_ref_t automaton, __tavr_auto_proc_t autoProc);
void		__tavr_add_post_layer		(void);
void		__tavr_remove_post_layer	(void);

void		__tavr_out_call				(void * subscribeList, __tavr_ref_t event);
void		__tavr_release_out_list		(void ** subscribeList);
void		__tavr_add_out_list			(void ** subscribeList, __tavr_ref_t automaton, __tavr_auto_proc_t autoProc, int isPost);
void		__tavr_remove_out_list		(void ** subscribeList, __tavr_ref_t automaton);

//////////////////////////////////////////////////////////////////////////
// Memory fuctions

#define __TAVR_ARRAY_OFFSET				3

#ifndef __TAVR_ENABLE_MEM_STAT
#define __TAVR_MEM_STAT_INFO
#define __TAVR_MEM_STAT_VALUE
#define __TAVR_MEM_STAT_FILE_LINE
#else
#define __TAVR_MEM_STAT_INFO , char * file, int line
#define __TAVR_MEM_STAT_VALUE , file, line
#define __TAVR_MEM_STAT_FILE_LINE	, __FILE__, __LINE__
#endif

void *		__tavr_alloc				(__tavr_int size, void * typeinfo __TAVR_MEM_STAT_INFO);
void		__tavr_free					(void * ptr				__TAVR_MEM_STAT_INFO);
void *		__tavr_alloc_array			(__tavr_int count, __tavr_int size, void * typeinfo __TAVR_MEM_STAT_INFO);
void		__tavr_free_array			(void * ptr				__TAVR_MEM_STAT_INFO);
void *		__tavr_alloc_string			(__tavr_char * value	__TAVR_MEM_STAT_INFO);
void *		__tavr_alloc_char_string	(char * value			__TAVR_MEM_STAT_INFO);

#ifdef __TAVR_ENABLE_MEM_STAT
#define __tavr_alloc(size, typeinfo)	__tavr_alloc(size, typeinfo, __FILE__, __LINE__)
#define __tavr_free(ptr)				__tavr_free(ptr, __FILE__, __LINE__)
#define __tavr_alloc_array(count, size, typeinfo) __tavr_alloc_array(count, size, typeinfo, __FILE__, __LINE__)
#define __tavr_free_array(ptr)			__tavr_free_array(ptr, __FILE__, __LINE__)
#define __tavr_alloc_string(value)		__tavr_alloc_string(value, __FILE__, __LINE__)
#endif

void		__tavr_assert				(char * cond, char * filename, int line);

#ifdef __cplusplus
} // extern "C"
#endif

#endif // _TAVR_RUNTIME_H_
