
#ifndef TAVR_RUNTIME_H
#define TAVR_RUNTIME_H

#pragma once

#include "tavr_runtime.h"

#undef __tavr_alloc
#undef __tavr_free
#undef __tavr_alloc_array
#undef __tavr_free_array
#undef __tavr_alloc_string

#pragma warning(disable : 4996)
#pragma warning(disable : 4311)

#include <stdlib.h>
#include <stdio.h>
#include <set>
#include <queue>

#include <ecg\ecg_comm.h>

namespace tavr {

template <class T>
class REF {
public:
	T *		operator ->		(void) { return ptr; }

	int		refcount		(void) { return ref[0]; }

	REF(void * data)
	{
		ref = (int *)data;
		ptr = (T *)(ref + 2);
	}

	REF(const REF & p)
	{
		ref = p.ref;
		ptr = p.ptr;
		++ref[0];
	}

	REF(__tavr_ref_t p)
	{
		ref = p.ref;
		ptr = p.ptr;
		++ref[0];
	}

	void operator = (const REF & p)
	{
		this->~REF();
		new (this) REF(p);
	}

	void operator = (__tavr_ref_t p)
	{
		~REF();
		new (this) REF(p);
	}

	operator __tavr_ref_t (void)
	{
		++ref[0];
		__tavr_ref_t res = { ref, ptr };
		return res;
	}

	~REF(void)
	{
		if (ref == NULL) {
			return ;
		}
		if (!--ref[0]) {
			ptr->~T();
			__tavr_free(ref __TAVR_MEM_STAT_FILE_LINE);
		}
	}

private:
	int *	ref;
	T *		ptr;
};

template <class T>
class REF_ARRAY {
public:
	int			length(void) { return ref[2]; }
	int			refcount(void) { return ref[0]; }
	T &			operator [] (int index) { return ptr[index]; }

	REF_ARRAY(void * data)
	{
		ref = (int *)data;
		ptr = (T *)(ref + 3);
	}

	REF_ARRAY(const REF_ARRAY & p)
	{
		ref = p.ref;
		ptr = p.ptr;
		++ref[0];
	}

	REF_ARRAY(__tavr_ref_t p)
	{
		ref = p.ref;
		ptr = p.ptr;
		++ref[0];
	}

	void operator = (const REF_ARRAY & p)
	{
		~REF_ARRAY();
		new (this) REF_ARRAY(p);
	}

	void operator = (__tavr_ref_t p)
	{
		~REF_ARRAY();
		new (this) REF_ARRAY(p);
	}

	operator __tavr_ref_t (void)
	{
		++ref[0];
		__tavr_ref_t res = { ref, ptr };
		return res;
	}

	~REF_ARRAY(void)
	{
		if (ref == NULL) {
			return ;
		}
		if (!--ref[0]) {
			for (int i=0; i<ref[2]; ++i) {
				(*this)[i].~T();
			}
			__tavr_free_array(ref __TAVR_MEM_STAT_FILE_LINE);
		}
	}

private:
	int *		ref;
	T *			ptr;
};

struct STRING {
	int			length;
	wchar_t		value[1];
};

typedef REF<STRING> REF_STRING;

//////////////////////////////////////////////////////////////////////////
// Allocation functions

#ifdef __TAVR_ENABLE_MEM_STAT
struct STAT_INFO {
	void *			ptr;
	int				size;
	const char *	file;
	int				line;

	bool operator < (const STAT_INFO & p) const
	{
		return ptr < p.ptr;
	}
};

typedef std::set<STAT_INFO> STAT_INFO_SET;
extern STAT_INFO_SET memStatInfo;
#endif

/*************************************************************************/
/* * Alloc()                                                             */
/*                                                                       */
/*************************************************************************/
inline void * Alloc(__tavr_int size __TAVR_MEM_STAT_INFO)
{
	void * res = malloc(size);
#ifdef PRINT_ALLOCATIONS
	printf("alloc 0x%08x size=%d\n", (int)res, size);
#endif
#ifdef __TAVR_ENABLE_MEM_STAT
	STAT_INFO info;
	info.ptr		= res;
	info.size		= size;
	info.file		= file;
	info.line		= line;
	__TAVR_ASSERT(memStatInfo.find(info) == memStatInfo.end(), "Internal error");
	memStatInfo.insert(info);
#endif
	return res;
}

/*************************************************************************/
/* Free()                                                                */
/*                                                                       */
/*************************************************************************/
inline void Free(void * ptr __TAVR_MEM_STAT_INFO)
{
#ifdef PRINT_ALLOCATIONS
	printf("free 0x%08x\n", (int)ptr);
#endif
#ifdef __TAVR_ENABLE_MEM_STAT
	STAT_INFO info;
	info.ptr = ptr;
	__TAVR_ASSERT(memStatInfo.find(info) != memStatInfo.end(), "Internal error");
	memStatInfo.erase(info);
#endif
	free(ptr);
}

//////////////////////////////////////////////////////////////////////////
// Event data
struct PENDING_EVENT {
	__tavr_ref_t		event;
	__tavr_ref_t		dstAuto;
	__tavr_auto_proc_t	proc;
};
typedef std::queue<PENDING_EVENT> PENDING_QUEUE;

struct SUBSCRIBED_AUTO {
	__tavr_ref_t		srcAuto;
	__tavr_auto_proc_t	proc;
	int					used;
	bool				isSend;

	SUBSCRIBED_AUTO *	next;
};

} // namespace tavr

extern "C" {

extern void * __tavr_ti_string[];
extern void * __tavr_tirt_ref_array[];
int __tavr__main___string_a(__tavr_ref_t);

}

#ifdef __TAVR_ENABLE_MEM_STAT
#define TAVR_NEW new (__FILE__, __LINE__)
#define TAVR_DELETE(PTR) tavr::Free((PTR), __FILE__, __LINE__)

inline void * operator new (size_t size, char * file, int line)
{ return tavr::Alloc((int)size, file, line); }

inline void operator delete (void * ptr, char * file, int line)
{ tavr::Free(ptr, file, line); }

#else
#define TAVR_NEW new
#endif

#define TAVR_INC_REF __TAVR_INC_REF

#define TAVR_DEC_REF(REF) \
	if ((REF).ref) { \
		/*printf("dec_ref(%d) 0x%08x (%s, %d)\n", (REF).ref[0], (int)(REF).ref, __FILE__, __LINE__);*/ \
		if (!--(*(REF).ref)) { \
			((__tavr_destr_t)((void **)(REF).ref[1])[1])(&REF); \
			__tavr_free((REF).ref __TAVR_MEM_STAT_FILE_LINE); \
		} \
	}


#endif // TAVR_RUNTIME_H