
#include "stdafx.h"
#include "tavr.h"
#include "tavr_target_c.h"
#include "ecg\misc\ecg_os.h"

using namespace std;
using namespace ecg;

namespace tavr {

struct TARGET_C::TMP_VAR {
	int							index;
	TYPEINFO					type;
	int							argSet;

	TMP_VAR(void)
		: index(-1), argSet(-1)
	{}

	inline bool operator < (const TMP_VAR & tmp) const
	{
		if (type < tmp.type) {
			return true;
		}
		if (tmp.type < type) {
			return false;
		}
		if (index == -1 || tmp.index == -1) {
			return false;
		}
		return index < tmp.index;
	}
};

TAVR_TYPE_SPECIAL_NAMES_T TAVR_TYPE_SPECIAL_NAMES[] = {
	{ "=",		"op_asgn" },
	{ "(T)",	"op_" },
};

/************************************************************************/
/* tavr::TARGET_C::INFO::INFO
/*
/************************************************************************/
TARGET_C::INFO::INFO(void)
	: blocksCounter(), curProc()
{}

/************************************************************************/
/* tavr::TARGET_C::GetRefName
/*
/************************************************************************/
std::string TARGET_C::GetRefName(const TYPEINFO & ti)
{
	TYPEINFO n;
	n.type		= TYPEINFO::TYPE_TYPE;
	n.info.type = ti.info.type;

	string res;
	if (ti.array == 0) {
		res = TAVR_REF;
	} else {
		n.array = ti.array - 1;
		res = TAVR_REF_ARRAY;
	}

	res += GetTypeName(n, true);
	return res;
}

/************************************************************************/
/* tavr::TARGET_C::GetDecoratedName
/*
/************************************************************************/
std::string TARGET_C::GetDecoratedName(const std::string & name)
{
	string res;
	res.reserve(name.size());

	for (int i=0, size = (int)name.size(); i < size; ++i) {
		if (name[i] != '_') {
			res += name[i];
		} else {
			res += "__";
		}
	}

	return res;
}

/************************************************************************/
/* tavr::TARGET_C::GetTypeName
/*
/************************************************************************/
std::string TARGET_C::GetTypeName(const TYPEINFO & ti, bool isDecoratedName /* = false */)
{
	assert(ti.type == TYPEINFO::TYPE_TYPE || ti.IsValue());
	assert(ti.info.type != NULL);

	int ptrCount = ti.isRef + ti.array;

	string res;
	res.reserve(128);

	TYPEINFO userType = ti;
	userType.array = 0;
	const bool isUserType = userType.IsUserType();
	
	if (ti.IsNull()) {
		if (!isDecoratedName) {
			return TAVR_NULL_TYPE;
		}
		return TAVR_NULL_TYPE_DEC;
	}

	if (!isDecoratedName) {
		if (isUserType || ti.array > 0) {
			res += "struct ";
		} else if (ti.info.type != project->defType[PROJECT::DEF_TYPE_VOID]) {
			res += TAVR_PREFIX;
		}
	}

	if (isDecoratedName || !ti.IsTypeByRef()) {
		res += GetDecoratedName(ti.info.type->name.str());
	} else {
		res += GetRefName(ti);
	}

	if (!isDecoratedName) {
		if (ti.isRef) {
			res += "*";
		}
	} else if (ti.isRef || ti.array > 0) {
		res += "_";
		if (ti.isRef) {
			res += TAVR_REF_POSTFIX;
		}

		if (ti.array > 0) {
			res += TAVR_ARRAY_POSTFIX;
			if (ti.array > 1) {
				res += ToString(ti.array);
			}
		}
	}

	return res;
}

/************************************************************************/
/* tavr::TARGET_C::GetProcName
/*
/************************************************************************/
std::string TARGET_C::GetProcName(METHOD * proc)
{
	assert(proc != NULL);

	string res = TAVR_PREFIX "_";
	TYPEINFO ti;
	if (proc->IsMethod()) {
		ti.type = TYPEINFO::TYPE_TYPE;
		ti.info.type = proc->owner.type;
		res += GetTypeName(ti, true);
		res += "___";
	}

	switch (proc->type) {
		case METHOD::TYPE_CONSTRUCTOR:
			if (!proc->owner.type->attribs.is(ATTRIB_EXTERN)) {
				res += GetTypeName(ti, true);
			} else {
				res += TAVR_NEW;
			}
			break;

		case METHOD::TYPE_DESTRUCTOR:
			res += TAVR_DESTRUCTOR "_";
			res += GetTypeName(ti, true);
			break;

		case METHOD::TYPE_OPERATOR:
			{
				bool found = false;
				
				for (int i=0; i<ECG_NUM_ELEMENTS(TAVR_TYPE_SPECIAL_NAMES); ++i) {
					if (proc->name.str() == TAVR_TYPE_SPECIAL_NAMES[i].name) {
						res += TAVR_TYPE_SPECIAL_NAMES[i].decoration;
						found = true;
						break;
					}
				}
				assert(found);

				if (proc->name == project->castOpString) {
					res += GetTypeName(proc->result, true);
				}
			}
			break;

		case METHOD::TYPE_AUTO_PROC:
			{
				TYPE * type = proc->owner.type;
				res += project->autoString.str();
				if (proc == type->specs[TYPE::SPEC_AUTO_PROC]) {
					assert(proc->name == project->autoString);
				} else {
					bool found = false;
					for (int i=0; i<(int)type->subautos.size(); ++i) {
						if (type->subautos[i].proc == proc) {
							res += "_";
							res += ToString(type->subautos[i].id);
							found = true;
							break;
						}
					}
					assert(found);
				}
			}
			break;

		default:
			res += GetDecoratedName(proc->name.str());
			break;
	}

	for(int i=0; i<(int)proc->args.size(); ++i) {
		res += "___";
		ti = proc->args[i].type;
		res += GetTypeName(ti, true);
	}

	return res;
}

/************************************************************************/
/* tavr::TARGET_C::PrintAssignOp
/*
/************************************************************************/
void TARGET_C::PrintAssignOp(bool thisCall, const TYPEINFO & leftType, const std::string & left, COMMAND_INFO & cmdInfo, INFO & localInfo, const EVAL_COMMAND * cmd, const EVAL_CODE * code, std::string * right /*= NULL*/)
{
	METHOD * fieldOpAssign;
	if (cmd == NULL || cmd->info.proc == NULL) {
		fieldOpAssign = leftType.info.type->specs[TYPE::SPEC_DEF_OP_ASSIGN];
	} else {
		assert(leftType.IsTypeByValue());
		fieldOpAssign = cmd->info.proc;
	}

	assert(cmd != NULL || right != NULL);

	if (leftType.IsTypeByValue() && fieldOpAssign != NULL) {
		string argsVal = ProcessProcArgs(thisCall, fieldOpAssign, cmdInfo, localInfo, cmd, code, right);

		output << tabs << GetProcName(fieldOpAssign) << "(&" << left << ", " << argsVal << ");" << endl;
	} else{
		const bool isRef = leftType.IsTypeByRef();

		if (isRef) {
			PrintDestructor(leftType, left);
		}

		if (cmd != NULL) {
			output << tabs << left << " = " << localInfo.cmdInfo[cmd->args[thisCall]].value << ";" << endl;
		} else {
			output << tabs << left << " = " << *right << ";" << endl;
		}

		if (isRef) {
			output << tabs << TAVR_INC_REF "(" << left << ");" << endl;
		}
	}
}

/************************************************************************/
/* tavr::TARGET_C::PrintCopyConstr
/*
/************************************************************************/
void TARGET_C::PrintCopyConstr(const TYPEINFO & leftType, const std::string & left, const std::string & right)
{
	METHOD * copyConstr = leftType.info.type->specs[TYPE::SPEC_COPY_CONSTR];
	if (leftType.IsTypeByValue() && copyConstr != NULL) {
		output << tabs << GetProcName(copyConstr) << "(&" << left << ", ";
		assert(copyConstr->args[0].type.isRef);
		output << "&" << right << ");" << endl;
	} else if (!leftType.IsNull()) {
		output << tabs << left << " = " << right << ";" << endl;
		if (leftType.IsTypeByRef()) {
			output << tabs << TAVR_INC_REF "(" << left << ");" << endl;
		}
	} else {
		output << tabs << left << " = *(" TAVR_NULL_TYPE "*)&" << right << ";" << endl;
		output << tabs << TAVR_INC_REF "(" << left << ");" << endl;
	}
}

/************************************************************************/
/* tavr::TARGET_C::Generate
/*
/************************************************************************/
bool TARGET_C::Generate(MODULE * module, std::string & outputName)
{
	outputName = tmp + "\\" + module->name + ".c";

	// TODO: support project settings (temp dir and so on)
	output.open(outputName.c_str());

	if (!output) {
		return false;
	}

	output << "#include <tavr_runtime.h>" << endl;

	// meta processing
	std::vector<INFO> info;
	info.reserve(64);

	GLOBAL_INFO globalInfo;

	for (MODULE_MEMBER * mem = module->members; mem != NULL; mem = mem->next) {
		info.push_back(INFO());

		switch (mem->type) {
			case MODULE_MEMBER::TYPE_METHOD:
				PreProcessProc(mem->info.proc, info.back(), globalInfo);
				break;

			case MODULE_MEMBER::TYPE_TYPE:
				PreProcessType(mem->info.type, info.back(), globalInfo);
				break;

			default:
				assert(!"Pregenerate code: unknown module member");
				break;
		}
	}

	int index = 0;
	for (MODULE_MEMBER * mem = module->members; mem != NULL; mem = mem->next, ++index) {
		switch (mem->type) {
			case MODULE_MEMBER::TYPE_METHOD:
				ProcessProc(mem->info.proc, info[index], globalInfo);
				break;

			case MODULE_MEMBER::TYPE_TYPE:
				ProcessType(mem->info.type, info[index], globalInfo);
				break;

			default:
				assert(!"Generate code: unknown module member");
				break;
		}
	}


	output.close();

	// store generated forward declarations (for typeinfo)
	for (std::set<TYPEINFO>::const_iterator iter = globalInfo.forwardTypes.begin(); iter != globalInfo.forwardTypes.end(); ++iter) {
		typeinfoTypes.insert(*iter);
	}

	moduleList.push_back(module);

	return true;
}

/*************************************************************************/
/* TARGET_C::PrintProcHeaderArrDestr()                                   */
/*                                                                       */
/*************************************************************************/
void TARGET_C::PrintProcHeaderArrDestr(const TYPEINFO & ti)
{
	output << "void " TAVR_ARRAY_DESTR << GetTypeName(ti, true)
		<< "(" << GetTypeName(ti) << " * " TAVR_THIS << ")";
}

/************************************************************************/
/* tavr::TARGET_C::PrintProcHeader
/*
/************************************************************************/
void TARGET_C::PrintProcHeader(METHOD * proc, GLOBAL_INFO & globalInfo)
{
	const bool isConstructor = (proc->type == METHOD::TYPE_CONSTRUCTOR);

	for (int i=(int)proc->args.size() - 1; i >= 0; --i) {
		PreProcessTypePrintDecl(proc->args[i].type, globalInfo);
	}

	PreProcessTypePrintDecl(proc->result, globalInfo);

	const bool isExternAutoConstr = (proc->type == METHOD::TYPE_CONSTRUCTOR
		&& proc->owner.type->attribs.is(ATTRIB_EXTERN)
		&& (proc->owner.type->subtype == TYPE::SUBTYPE_AUTO
		|| proc->owner.type->subtype == TYPE::SUBTYPE_EVENT));

	assert(!isExternAutoConstr || proc->owner.type->subtype == TYPE::SUBTYPE_AUTO);

	if (!isExternAutoConstr) {
		output << tabs << GetTypeName(proc->result) << " ";
	} else {
		TYPEINFO ti;
		ti.SetType(project->defType[PROJECT::DEF_TYPE_VOID]);
		ti.isRef = true;

		output << tabs << GetTypeName(ti) << " ";
	}

	output << GetProcName(proc) << "(";

	if (proc->IsMethod() && !proc->attribs.is(ATTRIB_STATIC) && !isExternAutoConstr) {
		TYPEINFO ti;
		ti.type = TYPEINFO::TYPE_TYPE;
		ti.info.type = proc->owner.type;
		ti.isRef = true;
		output << GetTypeName(ti) << " " TAVR_THIS;
		if (!proc->args.empty()) {
			output << ", ";
		}
	}

	for (int j=0; j<(int)proc->args.size(); ++j) {
		const METHOD::ARG & arg = proc->args[j];

		output << GetTypeName(arg.type);

		if (arg.name != project->emptyString) {
			output << " " << arg.name.c_str();
		} else {
			const bool isDefCopyConstr = (proc->owner.type->specs[TYPE::SPEC_COPY_CONSTR]	== proc && proc->impl == NULL);
			const bool isDefAssignOp   = (proc->owner.type->specs[TYPE::SPEC_DEF_OP_ASSIGN]	== proc && proc->impl == NULL);

			if (isDefCopyConstr || isDefAssignOp) {
				output << " " TAVR_SRC;
			} else {
				output << " " TAVR_DUMMY_ARG << ToString(j);
			}
		}

		if (j + 1 < (int)proc->args.size()) {
			output << ", ";
		}
	}
	output << ")";
}

/************************************************************************/
/* tavr::TARGET_C::PrintVarDefInit
/*
/************************************************************************/
void TARGET_C::PrintVarDefInit(const TYPEINFO & type, const std::string & value)
{
	if (type.IsTypeByValue()) {
		// call default constructor
		output << tabs << GetProcName(type.info.type->specs[TYPE::SPEC_DEF_CONSTR])
			<< "(&" << value << ");" << endl;
	} else if (type.IsTypeByRef()) {
		output << tabs << value << ".ptr = 0;" << endl;
		output << tabs << value << ".ref = 0;" << endl;
	} else {
		output << tabs << value << " = 0;" << endl;
	}
}

/************************************************************************/
/* tavr::TARGET_C::PreProcessTypePrintDecl
/*
/************************************************************************/
void TARGET_C::PreProcessTypePrintDecl(const TYPEINFO & ti, GLOBAL_INFO & globalInfo)
{
	// ignore integrated types
	if (!ti.IsTypeByValue() && !ti.IsTypeByRef()) {
		return ;
	}

	TYPEINFO tmp = ti;
	tmp.type = TYPEINFO::TYPE_TYPE;
	tmp.isRef = false;

	if (globalInfo.forwardTypes.find(tmp) != globalInfo.forwardTypes.end()) {
		return ;
	}
	globalInfo.forwardTypes.insert(tmp);

	// type declaration
	if (ti.array == 0) {
		// field types
		for (int i=(int)ti.info.type->fields.size() - 1; i>=0; --i) {
			PreProcessTypePrintDecl(ti.info.type->fields[i].type, globalInfo);
		}

		TYPEINFO tmpTi = ti;
		tmpTi.isRef = false;
		output << "struct " << GetTypeName(tmpTi, true) << " {" << endl;

		// relax context value if needed (main auto proc)
		bool wasRelaxData = false;
		if (!ti.info.type->relaxInfo.context.empty()) {
			output << "	" TAVR_PREFIX "int " TAVR_RELAX_CONTEXT ";" << endl;
			wasRelaxData = true;
		}
		if (ti.info.type->relaxInfo.stackSize > 0) {
			output << "	" TAVR_PREFIX "byte " TAVR_RELAX_STACK "[" << ti.info.type->relaxInfo.stackSize << "];" << endl;
			wasRelaxData = true;
		}

		// generate relax context value for subautos
		for (int i=0; i < (int)ti.info.type->subautos.size(); ++i) {
			const SUBAUTO_INFO & si = ti.info.type->subautos[i];

			if (!si.relaxInfo.context.empty()) {
				output << "	" TAVR_PREFIX "int " TAVR_RELAX_CONTEXT "_" << si.id << ";" << endl;
				wasRelaxData = true;
			}
			if (si.relaxInfo.stackSize > 0) {
				output << "	" TAVR_PREFIX "byte " TAVR_RELAX_STACK "_" << si.id << "[" << si.relaxInfo.stackSize << "];" << endl;
				wasRelaxData = true;
			}
		}

		// data for out actions
		for (TYPE::OUT_TABLE::const_iterator iter = ti.info.type->outTable.begin()
			, iterEnd = ti.info.type->outTable.end()
			; iter != iterEnd
			; ++iter)
		{
			output << "	void * " << TAVR_OUT_PREFIX << iter->first.str() << ";" << endl;
		}

		for (int i=(int)ti.info.type->fields.size() - 1; i>=0; --i) {
			const TYPE::FIELD & field = ti.info.type->fields[i];
			output << "	" << GetTypeName(field.type) << " " << field.name.str() << ";" << endl;
		}

		// generate dummy value if needed (struct could not be empty)
		if (ti.info.type->fields.empty()
		&& !wasRelaxData
		&& ti.info.type->outTable.empty()) {
			output << "	" TAVR_PREFIX "int " TAVR_DUMMY << ToString(globalInfo.dummyCounter++) << ";" << endl;
		}
		output << "};" << endl;

		if (METHOD * destr = ti.info.type->specs[TYPE::SPEC_DESTR]) {
			PreProcessProcPrintHeader(destr, globalInfo);
		}
	}

	// type ptr
	if (ti.IsTypeByRef()) {
		output << "struct " << GetRefName(ti) << " {" << endl;
		output << "	" TAVR_PREFIX "int* ref;" << endl;
		TYPEINFO tmp = ti;
		if (ti.array > 0) {
			tmp.array = ti.array - 1;
			output << "	" << GetTypeName(tmp) << "* ptr;" << endl;
		} else {
			output << "	struct " << GetTypeName(ti, true) << "* ptr;" << endl;
		}
		output << "};" << endl;

		if (ti.array > 0) {
			PreProcessTypePrintDecl(tmp, globalInfo);
		}

		// typeinfo extern pointer
		output << "void * " TAVR_TYPEINFO << GetTypeName(ti, true) << "[];" << endl;
	}
}

/************************************************************************/
/* tavr::TARGET_C::PreProcessProcPrintHeader
/*
/************************************************************************/
void TARGET_C::PreProcessProcPrintHeader(METHOD * proc, GLOBAL_INFO & globalInfo)
{
	if (globalInfo.forwardProcs.find(proc) == globalInfo.forwardProcs.end()) {
		globalInfo.forwardProcs.insert(proc);

		if (proc->IsMethod()) {
			TYPEINFO ti;
			ti.type = TYPEINFO::TYPE_TYPE;
			ti.info.type = proc->owner.type;
			PreProcessTypePrintDecl(ti, globalInfo);
		}

		PrintProcHeader(proc, globalInfo);
		output << ";" << endl;
	}
}

/************************************************************************/
/* tavr::TARGET_C::PreProcessCodeBlockBegin
/*
/************************************************************************/
void TARGET_C::PreProcessCodeBlockBegin(INFO & localInfo, GLOBAL_INFO & globalInfo, COMMAND_INFO & cmdInfo)
{
	localInfo.argSets.push_back(INT_ARRAY());
	globalInfo.argSetIdx.push((int)localInfo.argSets.size() - 1);

	localInfo.tmpVarSets.push_back(TMP_VAR_ARRAY());

	cmdInfo.argSet = globalInfo.argSetIdx.top();
}

/************************************************************************/
/* tavr::TARGET_C::PreProcessCodeBlockBegin
/*
/************************************************************************/
void TARGET_C::PreProcessCodeBlockEnd(INFO & localInfo, GLOBAL_INFO & globalInfo, COMMAND_INFO & cmdInfo)
{
	cmdInfo.argSet = globalInfo.argSetIdx.top();
	globalInfo.argSetIdx.pop();

	// remove tmp vars that should be popped from unused list
	TMP_VAR_ARRAY & tmpVars = localInfo.tmpVarSets[cmdInfo.argSet];

	for (TMP_VAR_SET::iterator iter = localInfo.unusedTmpVars.begin(); iter != localInfo.unusedTmpVars.end(); ) {
		bool found = false;

		for (int i=(int)tmpVars.size() - 1; i >= 0; --i) {
			if (tmpVars[i].index == iter->index) {
				found = true;
				break;
			}
		}

		if (found) {
			iter = localInfo.unusedTmpVars.erase(iter);
		} else {
			++iter;
		}
	}
}

/************************************************************************/
/* tavr::TARGET_C::GetTmpVar
/*
/************************************************************************/
int TARGET_C::GetTmpVar(const TYPEINFO & ti, bool needDestructor, COMMAND_INFO & cmdInfo, INFO & localInfo, GLOBAL_INFO & globalInfo)
{
	if (ti.IsUserType()) {
		if (METHOD * destr = ti.info.type->specs[TYPE::SPEC_DESTR]) {
			PreProcessProcPrintHeader(destr, globalInfo);
		}
	}

	TMP_VAR var;
	var.type		= ti;
	var.type.type	= TYPEINFO::TYPE_TYPE;
	var.index		= -1;

	PENDING_TMP_VAR ptv;
	ptv.needDestructor	= needDestructor;

	TMP_VAR_SET::iterator varIter = localInfo.unusedTmpVars.find(var);
	if (varIter != localInfo.unusedTmpVars.end()) {
		ptv.index			= varIter->index;
		ptv.argSet			= varIter->argSet;

		localInfo.unusedTmpVars.erase(varIter);

		cmdInfo.pendingTmpVars.push_back(ptv);
		return ptv.index;
	}

	var.index	= globalInfo.tmpVarCounter++;
	var.argSet	= globalInfo.argSetIdx.top();

	localInfo.tmpVarSets[var.argSet].push_back(var);

	ptv.index	= var.index;
	ptv.argSet	= var.argSet;
	cmdInfo.pendingTmpVars.push_back(ptv);

	return var.index;
}

/************************************************************************/
/* tavr::TARGET_C::PreProcessProcArgs
/*
/************************************************************************/
void TARGET_C::PreProcessProcArgs(METHOD * proc, COMMAND_INFO & cmdInfo, INFO & localInfo, GLOBAL_INFO & globalInfo)
{
	assert(proc != NULL);

	assert(cmdInfo.tmpVars.empty());
	int numArgs = (int)proc->args.size();
	cmdInfo.tmpVars.resize(numArgs, -1);

	// tmp for assign operator argument
	for (int i=0; i<numArgs; ++i) {
		TYPEINFO & argType = proc->args[i].type;
		if (argType.isRef) {
			continue;
		}

		cmdInfo.tmpVars[i] = GetTmpVar(argType, false, cmdInfo, localInfo, globalInfo);

		if (METHOD * copyConstr = argType.info.type->specs[TYPE::SPEC_COPY_CONSTR]) {
			PreProcessProcPrintHeader(copyConstr, globalInfo);
		}
	}
}

/************************************************************************/
/* tavr::TARGET_C::PreProcessReleaseTmpVars
/*
/************************************************************************/
void TARGET_C::PreProcessReleaseTmpVars(PENDING_TMP_VAR_ARRAY & pendingTmpVars, COMMAND_INFO & cmdInfo, INFO & localInfo)
{
	// release pending tmp vars
	for (int i=(int)pendingTmpVars.size() - 1; i >= 0; --i) {
		PENDING_TMP_VAR & ptv = pendingTmpVars[i];
		TMP_VAR_ARRAY & tmpVars = localInfo.tmpVarSets[ptv.argSet];

		bool found = false;
		for (int k=(int)tmpVars.size() - 1; k >= 0; --k) {
			const TMP_VAR & tmpVar = tmpVars[k];

			if (tmpVar.index == ptv.index) {
				localInfo.unusedTmpVars.insert(tmpVar);
				found = true;
				break;
			}
		}
		assert(found);
	}
}

/************************************************************************/
/* tavr::TARGET_C::PreProcessCode
/*
/************************************************************************/
void TARGET_C::PreProcessCode(const EVAL_CODE & code, INFO & localInfo, GLOBAL_INFO & globalInfo)
{
	std::vector<COMMAND_INFO> & info = localInfo.cmdInfo;

	info.resize(code.size());

	PENDING_TMP_VAR_ARRAY pendingTmpVars;
	PENDING_TMP_VAR_ARRAY forceReleaseTmpVars;
	int curRelaxStackSize = 0;

	std::stack<int> curRelaxStackSizeHistory;

	// print forward declarations for used functions
	for (int i=0; i<(int)code.size(); ++i) {
		const EVAL_COMMAND &	cmd		= code[i];
		COMMAND_INFO &			cmdInfo	= info[i];

		if (cmd.type != EVAL_COMMAND::TYPE_VARIABLE) {
			for (int j=(int)cmd.args.size() - 1; j >= 0; --j) {
				swap(pendingTmpVars, info[cmd.args[j]].pendingTmpVars);
				copy(pendingTmpVars.begin(), pendingTmpVars.end(), back_inserter(cmdInfo.pendingTmpVars));
				pendingTmpVars.clear();
			}
		}

		if (!globalInfo.argSetIdx.empty()) {
			cmdInfo.argSet = globalInfo.argSetIdx.top();
		}

		switch (cmd.type) {
			case EVAL_COMMAND::TYPE_SIZEOF:
				{
					// this is a hack!
					EVAL_COMMAND & c = const_cast<EVAL_COMMAND &> (cmd);
					c.type		= EVAL_COMMAND::TYPE_CONSTANT;

					int size = TYPE_BY_REF_SIZE;
					if (!c.typeinfo.IsTypeByRef()) {
						size = c.typeinfo.info.type->size;
						assert(size > 0);
					}
					c.info.name = toSTR_ID(ToString(size));
					c.typeinfo.SetRValue(project->defType[PROJECT::DEF_TYPE_INT]);
				}
				break;

			case EVAL_COMMAND::TYPE_CONSTANT:
				if (cmd.typeinfo.info.type == project->defType[PROJECT::DEF_TYPE_STRING]) {
					cmdInfo.tmpVars.push_back(
						GetTmpVar(cmd.typeinfo, true, cmdInfo, localInfo, globalInfo));
				}
				break;

			case EVAL_COMMAND::TYPE_THIS_CALL:
			case EVAL_COMMAND::TYPE_PROC_CALL:
			case EVAL_COMMAND::TYPE_CONSTR_CALL:
				{
					METHOD * proc = cmd.info.proc;
					PreProcessProcPrintHeader(proc, globalInfo);

					const bool thisCall = (cmd.type == EVAL_COMMAND::TYPE_THIS_CALL);

					// check whether arguments need copy constructors
					PreProcessProcArgs(proc, cmdInfo, localInfo, globalInfo);

					// does result need tmp
					bool needTmpVar = false;
					TYPEINFO tmpVarType;
					if (proc->result.IsTypeByValue() || proc->result.IsTypeByRef()) {
						tmpVarType	= proc->result;
						needTmpVar	= true;
					} else if (cmd.type == EVAL_COMMAND::TYPE_CONSTR_CALL) {
						tmpVarType.SetType(proc->owner.type);
						needTmpVar	= true;
					}

					if (needTmpVar) {
						cmdInfo.tmpVars.push_back(GetTmpVar(tmpVarType, true, cmdInfo, localInfo, globalInfo));
					}
				}
				break;

			case EVAL_COMMAND::TYPE_REG_POST:
			case EVAL_COMMAND::TYPE_REG_SEND:
			case EVAL_COMMAND::TYPE_SEND:
			case EVAL_COMMAND::TYPE_POST:
				{
					assert(cmd.args.size() == 2);
					assert(cmd.info.proc != NULL && cmd.info.proc->type == METHOD::TYPE_AUTO_PROC);
					PreProcessProcPrintHeader(cmd.info.proc, globalInfo);
				}
				break;

			case EVAL_COMMAND::TYPE_CODE_BLOCK_BEGIN:
				PreProcessCodeBlockBegin(localInfo, globalInfo, cmdInfo);

				curRelaxStackSizeHistory.push(curRelaxStackSize);
				break;

			case EVAL_COMMAND::TYPE_CODE_BLOCK_END:
				PreProcessCodeBlockEnd(localInfo, globalInfo, cmdInfo);

				assert(!curRelaxStackSizeHistory.empty());
				curRelaxStackSize = curRelaxStackSizeHistory.top();
				curRelaxStackSizeHistory.pop();
				break;

			case EVAL_COMMAND::TYPE_DECL_RELAX:
				// compose value name for variable in relax stack
				{
					assert(localInfo.curProc != NULL);
					assert(localInfo.curProc->type == METHOD::TYPE_AUTO_PROC);
					TYPE * curType = localInfo.curProc->owner.type;

					SUBAUTO_INFO * subautoInfo = NULL;
					for (int i=0; i<(int)curType->subautos.size(); ++i) {
						if (localInfo.curProc == curType->subautos[i].proc) {
							subautoInfo = &curType->subautos[i];
							break;
						}
					}
					assert(subautoInfo != NULL || localInfo.curProc == curType->specs[TYPE::SPEC_AUTO_PROC]);

					string value = "(*(";
					value += GetTypeName(cmd.typeinfo);
					if (curType->IsTypeByRef()) {
						value += "*)(" TAVR_THIS "->ptr->" TAVR_RELAX_STACK;
					} else {
						value += "*)(" TAVR_THIS "->" TAVR_RELAX_STACK;
					}
					if (subautoInfo != NULL) {
						value += "_";
						value += ToString(subautoInfo->id);
					}

					value += "+";
					value += ToString(curRelaxStackSize);
					value += "))";

					swap(cmdInfo.value, value);

					int size = TYPE_BY_REF_SIZE;
					if (!cmd.typeinfo.IsTypeByRef()) {
						size = cmd.typeinfo.info.type->size;
						assert(size > 0);
					}
					curRelaxStackSize += size;
				}
				// no break

			case EVAL_COMMAND::TYPE_DECL:
				localInfo.argSets[
					globalInfo.argSetIdx.top()
				].push_back(i);

				PreProcessTypePrintDecl(cmd.typeinfo, globalInfo);

				if (cmd.typeinfo.IsTypeByValue()) {
					TYPE * type = cmd.typeinfo.info.type;

					if (cmd.info.proc == NULL) {
						METHOD * proc = NULL;
						METHOD * copyConstr = type->specs[TYPE::SPEC_COPY_CONSTR];
						if (!cmd.args.empty() && copyConstr != NULL) {
							// call copy constructor when we've got it and there is an initialization expression
							assert(cmd.args.size() == 1);
							proc = copyConstr;
						} else {
							// otherwise call default constructor first
							proc = type->specs[TYPE::SPEC_DEF_CONSTR];
						}

						PreProcessProcPrintHeader(proc, globalInfo);
						break;
					}

					// declaration with specific constructor
					assert(cmd.info.proc->type == METHOD::TYPE_CONSTRUCTOR);
					PreProcessProcArgs(cmd.info.proc, cmdInfo, localInfo, globalInfo);
						
					assert(cmd.info.proc != NULL);
					PreProcessProcPrintHeader(cmd.info.proc, globalInfo);
				}
				break;

			case EVAL_COMMAND::TYPE_IF:
				// need tmp var if there are pending tmp vars
				if (!cmdInfo.pendingTmpVars.empty()) {
					TYPEINFO ti;
					ti.SetType(project->defType[PROJECT::DEF_TYPE_BOOL]);
					cmdInfo.tmpVars.push_back(GetTmpVar(ti, false, cmdInfo, localInfo, globalInfo));
				}
				break;

			case EVAL_COMMAND::TYPE_EXPR_IF_COND:
				cmdInfo.tmpVars.push_back(GetTmpVar(cmd.typeinfo, false, cmdInfo, localInfo, globalInfo));
				break;

			case EVAL_COMMAND::TYPE_EXPR_IF_ELSE:
			case EVAL_COMMAND::TYPE_EXPR_IF_RES:
				{
					assert(cmd.args.size() >= 2);
					const int topIndex = cmd.args[0];

					for (int index = i - 1; index > topIndex; --index) {
						COMMAND_INFO & ci = info[index];
						for (int k=0; k<(int)ci.tmpVars.size(); ++k) {
							for (int l=0; l<(int)cmdInfo.pendingTmpVars.size(); ++l) {
								if (cmdInfo.pendingTmpVars[l].index != ci.tmpVars[k]) {
									continue;
								}
							
								forceReleaseTmpVars.push_back(cmdInfo.pendingTmpVars[l]);
								cmdInfo.pendingTmpVars.erase(cmdInfo.pendingTmpVars.begin() + l);
								break;
							}
						}
					}

					localInfo.forceReleaseTmpVars[i] = forceReleaseTmpVars;
				}
				break;

			case EVAL_COMMAND::TYPE_LABEL:
				if (cmd.isUsed) {
					cmdInfo.value = TAVR_LABEL + ToString(globalInfo.labelCounter++);
				}
				break;

			case EVAL_COMMAND::TYPE_NEW_ARRAY:
				{
					// count
					TYPEINFO ti;
					ti.type = TYPEINFO::TYPE_TYPE;
					ti.info.type = project->defType[PROJECT::DEF_TYPE_INT];
					cmdInfo.tmpVars.push_back(
						GetTmpVar(ti, false, cmdInfo, localInfo, globalInfo));

					// result
					cmdInfo.tmpVars.push_back(
						GetTmpVar(cmd.typeinfo, true, cmdInfo, localInfo, globalInfo));

					PreProcessTypePrintDecl(cmd.typeinfo, globalInfo);

					// forward declaration for constructor
					TYPEINFO elemType = cmd.typeinfo;
					assert(cmd.typeinfo.array > 0);
					--elemType.array;

					PreProcessTypePrintDecl(elemType, globalInfo);

					if (elemType.IsTypeByValue()) {
						TYPE * type = cmd.typeinfo.info.type;
						PreProcessProcPrintHeader(type->specs[TYPE::SPEC_DEF_CONSTR], globalInfo);
					}
				}
				break;

			case EVAL_COMMAND::TYPE_NEW_AUTO:
				{
					PreProcessProcArgs(cmd.info.proc, cmdInfo, localInfo, globalInfo);

					// result
					info[i].tmpVars.push_back(
						GetTmpVar(cmd.typeinfo, true, cmdInfo, localInfo, globalInfo));

					PreProcessTypePrintDecl(cmd.typeinfo, globalInfo);
					PreProcessProcPrintHeader(cmd.info.proc, globalInfo);
				}
				break;

			case EVAL_COMMAND::TYPE_ASSIGN:
				{
					if (cmd.typeinfo.IsTypeByValue()) {
						if (cmd.info.proc != NULL) {
							PreProcessProcArgs(cmd.info.proc, cmdInfo, localInfo, globalInfo);
						} else if (METHOD * defOpAssign = cmd.typeinfo.info.type->specs[TYPE::SPEC_DEF_OP_ASSIGN]) {
							PreProcessProcArgs(defOpAssign, cmdInfo, localInfo, globalInfo);
						}
					}

					// tmp for result type
					const TYPEINFO & rightType = code[cmd.args[1]].typeinfo;

					if ((cmd.isUsed || cmd.typeinfo.IsTypeByRef()) && !rightType.IsNull()) {
						// tmp for increment
						TYPEINFO ti = cmd.typeinfo;
						ti.isRef = true;
						cmdInfo.tmpVars.push_back(GetTmpVar(ti, false, cmdInfo, localInfo, globalInfo));
					}

					if (cmd.typeinfo.IsTypeByValue()) {
						METHOD * operatorAssign;
						if (cmd.info.proc == NULL) {
							operatorAssign = cmd.typeinfo.info.type->specs[TYPE::SPEC_DEF_OP_ASSIGN];
						} else {
							operatorAssign = cmd.info.proc;
						}
						if (cmd.typeinfo.IsTypeByValue() && operatorAssign != NULL) {
							PreProcessProcPrintHeader(operatorAssign, globalInfo);
						}
					}
				}
				break;

			case EVAL_COMMAND::TYPE_ARRAY_LENGTH:
				{
					assert(cmd.args.size() == 1);

					// array
					TYPEINFO ti = code[cmd.args[0]].typeinfo;
					cmdInfo.tmpVars.push_back(
						GetTmpVar(ti, false, cmdInfo, localInfo, globalInfo));
				}
				break;

			case EVAL_COMMAND::TYPE_STRING_INDEX:
			case EVAL_COMMAND::TYPE_ARRAY_INDEX:
				{
					assert(cmd.args.size() == 2);

					// array
					TYPEINFO ti = code[cmd.args[0]].typeinfo;

					cmdInfo.tmpVars.push_back(
						GetTmpVar(ti, false, cmdInfo, localInfo, globalInfo));

					// index
					ti.SetType(project->defType[PROJECT::DEF_TYPE_INT]);
					cmdInfo.tmpVars.push_back(
						GetTmpVar(ti, false, cmdInfo, localInfo, globalInfo));
				}
				break;

			case EVAL_COMMAND::TYPE_SUBNAME:
				{
					assert(cmd.args.size() == 1);
					const TYPEINFO & ti = code[cmd.args[0]].typeinfo;
					if (ti.IsTypeByRef()) {
						// tmp value for null ptr checking
						cmdInfo.tmpVars.push_back(
							GetTmpVar(ti, false, cmdInfo, localInfo, globalInfo));
					}
				}
				break;

			case EVAL_COMMAND::TYPE_RETURN:
				{
					if (!cmd.args.empty()) {
						assert(cmd.args.size() == 1);

						bool needTmp = cmd.typeinfo.IsTypeByRef();

						METHOD * copyConstr = cmd.typeinfo.info.type->specs[TYPE::SPEC_COPY_CONSTR];
						if (cmd.typeinfo.IsTypeByValue() && copyConstr != NULL) {
							PreProcessProcPrintHeader(copyConstr, globalInfo);
							needTmp = true;
						}

						if (needTmp) {
							cmdInfo.tmpVars.push_back(
								GetTmpVar(cmd.typeinfo, false, cmdInfo, localInfo, globalInfo));
						}
					}
				}
				break;
		}

		// tmp vars finalization
		if (!cmd.isUsed && !cmdInfo.pendingTmpVars.empty()) {
			PreProcessReleaseTmpVars(cmdInfo.pendingTmpVars, cmdInfo, localInfo);
		}
		if (!forceReleaseTmpVars.empty()) {
			PreProcessReleaseTmpVars(forceReleaseTmpVars, cmdInfo, localInfo);
			forceReleaseTmpVars.clear();
		}
	}
}

/*************************************************************************/
/* TARGET_C::PreProcessProcRelax()                                       */
/*                                                                       */
/*************************************************************************/
void TARGET_C::PreProcessProcRelax(METHOD * proc, RELAX_INFO & relaxInfo, COMMAND_INFO & cmdInfo, INFO & localInfo, GLOBAL_INFO & globalInfo)
{
	PreProcessCodeBlockBegin(localInfo, globalInfo, localInfo.defAssignOpInfo);

	// preprocess variables in relax stack
	const EVAL_CODE & code = proc->code;

	localInfo.relaxCopyTmpVars.resize(relaxInfo.context.size() + 1);

	for (map<int,int>::const_iterator iter = relaxInfo.context.begin(); iter != relaxInfo.context.end(); ++iter) {
		const EVAL_COMMAND & labelCmd = code[iter->first];
		assert(labelCmd.type == EVAL_COMMAND::TYPE_LABEL);

		INT_ARRAY & tmpVars = localInfo.relaxCopyTmpVars[iter->second];
		tmpVars.resize(labelCmd.availVars.size());

		for (int i=0; i<(int)labelCmd.availVars.size(); ++i) {
			const EVAL_COMMAND & varCmd = code[labelCmd.availVars[i]];
			assert(varCmd.type == EVAL_COMMAND::TYPE_DECL_RELAX);

			PreProcessFieldCopy(varCmd.typeinfo, localInfo.defAssignOpInfo, tmpVars[i], localInfo, globalInfo);
		}
	}

	PreProcessCodeBlockEnd(localInfo, globalInfo, localInfo.defAssignOpInfo);
}

/************************************************************************/
/* tavr::TARGET_C::PreProcessProc
/*
/************************************************************************/
void TARGET_C::PreProcessProc(METHOD * proc, INFO & localInfo, GLOBAL_INFO & globalInfo)
{
	localInfo.curProc = proc;

	bool isAssignOp = false;
	TYPE * type = NULL;

	// ignore constructors
	switch (proc->type) {
		case METHOD::TYPE_CONSTRUCTOR:
		case METHOD::TYPE_DESTRUCTOR:
		case METHOD::TYPE_OUT:
			return ;
		case METHOD::TYPE_OPERATOR:
			// skip generated by default operators
			if (proc->impl == NULL) {
				return ;
			}

			isAssignOp = proc->owner.type != NULL
				&& proc->owner.type->specs[TYPE::SPEC_DEF_OP_ASSIGN] == proc;

			type = proc->owner.type;
			break;
	}

	if (isAssignOp) {
		if (!type->relaxInfo.context.empty()) {
			assert(type->specs[TYPE::SPEC_AUTO_PROC] != NULL);
			PreProcessProcRelax(type->specs[TYPE::SPEC_AUTO_PROC], type->relaxInfo, localInfo.defAssignOpInfo, localInfo, globalInfo);
		}

		for (int i=0; i < (int)type->subautos.size(); ++i) {
			SUBAUTO_INFO & si = type->subautos[i];
			PreProcessProcRelax(si.proc, si.relaxInfo, localInfo.defAssignOpInfo, localInfo, globalInfo);
		}
	}

	// forward declaration for extern procedures
	if (proc->attribs.is(ATTRIB_EXTERN) || proc->type == METHOD::TYPE_AUTO_PROC) {
		PreProcessProcPrintHeader(proc, globalInfo);
	} else {
		assert(proc->impl != NULL);
	}

	PreProcessCode(proc->code, localInfo, globalInfo);
}

/************************************************************************/
/* tavr::TARGET_C::PreProcessFieldCopy
/*
/************************************************************************/
void TARGET_C::PreProcessFieldCopy(const TYPEINFO & ti, COMMAND_INFO & cmdInfo, int & tmpVar, INFO & localInfo, GLOBAL_INFO & globalInfo)
{
	METHOD * fieldOpAssign = ti.info.type->specs[TYPE::SPEC_DEF_OP_ASSIGN];
	if (ti.IsTypeByValue() && fieldOpAssign != NULL) {
		PreProcessProcArgs(fieldOpAssign, cmdInfo, localInfo, globalInfo);

		assert(cmdInfo.tmpVars.size() == 1);

		if (cmdInfo.tmpVars[0] != -1) {
			if (ti.info.type->specs[TYPE::SPEC_COPY_CONSTR] != NULL) {
				PreProcessProcPrintHeader(ti.info.type->specs[TYPE::SPEC_COPY_CONSTR], globalInfo);
			}
			PreProcessProcPrintHeader(fieldOpAssign, globalInfo);

			PreProcessReleaseTmpVars(cmdInfo.pendingTmpVars, cmdInfo, localInfo);
			assert(!cmdInfo.pendingTmpVars[0].needDestructor);
			cmdInfo.pendingTmpVars.clear();

			tmpVar = cmdInfo.tmpVars[0];
			cmdInfo.tmpVars.clear();
		}
	}
}

/************************************************************************/
/* tavr::TARGET_C::PreProcessType
/*
/************************************************************************/
void TARGET_C::PreProcessType(TYPE * type, INFO & info, GLOBAL_INFO & globalInfo)
{
	TYPEINFO ti;
	ti.type = TYPEINFO::TYPE_TYPE;
	ti.info.type = type;
	PreProcessTypePrintDecl(ti, globalInfo);

	info.fieldInfo.resize(type->fields.size());

	// emulate code block
	PreProcessCodeBlockBegin(info, globalInfo, info.constrCmdInfo);

	for (int i=0; i<(int)type->fields.size(); ++i) {
		const TYPE::FIELD & field = type->fields[i];

		// print def and copy constructors headers
		if (field.type.IsTypeByValue()) {
			TYPE * type = field.type.info.type;

			if (METHOD * forwConstr = type->specs[TYPE::SPEC_DEF_CONSTR]) {
				PreProcessProcPrintHeader(forwConstr, globalInfo);
			}
			if (METHOD * copyConstr = type->specs[TYPE::SPEC_COPY_CONSTR]) {
				PreProcessProcPrintHeader(copyConstr, globalInfo);
			}
			if (METHOD * forwDestr = type->specs[TYPE::SPEC_DESTR]) {
				PreProcessProcPrintHeader(forwDestr, globalInfo);
			}
			if (METHOD * copyOp = type->specs[TYPE::SPEC_DEF_OP_ASSIGN]) {
				PreProcessProcPrintHeader(copyOp, globalInfo);
			}
		}

		if (field.value != NULL) {
			swap(info.argSets,		info.fieldInfo[i].argSets);
			swap(info.tmpVarSets,	info.fieldInfo[i].tmpVarSets);

			PreProcessCode(field.code, info.fieldInfo[i], globalInfo);

			swap(info.argSets,		info.fieldInfo[i].argSets);
			swap(info.tmpVarSets,	info.fieldInfo[i].tmpVarSets);
		}
	}

	PreProcessCodeBlockEnd(info, globalInfo, info.constrCmdInfo);

	// process constructors code
	info.constrInfo.resize(type->methods.size());
	for (int i=0; i<(int)type->methods.size(); ++i) {
		METHOD * method = type->methods[i];
		if (method->type != METHOD::TYPE_CONSTRUCTOR) {
			continue;
		}

		// initialize argument set for default copy constructor
		if (method == type->specs[TYPE::SPEC_COPY_CONSTR] && method->impl == NULL) {
			PreProcessCodeBlockBegin(info, globalInfo, info.copyConstrCmdInfo);
			PreProcessCodeBlockEnd(info, globalInfo, info.copyConstrCmdInfo);
		}

		if (method == type->specs[TYPE::SPEC_EVENT_CONSTR] && method->impl == NULL) {
			PreProcessCodeBlockBegin(info, globalInfo, info.eventConstrCmdInfo);
			PreProcessCodeBlockEnd(info, globalInfo, info.eventConstrCmdInfo);
		}

		PreProcessCode(method->code, info.constrInfo[i], globalInfo);
	}

	// prepare copy constructors in relax stack
	if (!type->relaxInfo.context.empty()) {
		assert(type->specs[TYPE::SPEC_AUTO_PROC] != NULL);
		const EVAL_CODE & code = type->specs[TYPE::SPEC_AUTO_PROC]->code;

		for (map<int,int>::const_iterator iter = type->relaxInfo.context.begin(); iter != type->relaxInfo.context.end(); ++iter) {
			const EVAL_COMMAND & labelCmd = code[iter->first];

			for (int i=0; i<(int)labelCmd.availVars.size(); ++i) {
				const EVAL_COMMAND & varCmd = code[labelCmd.availVars[i]];

				TYPE * type = varCmd.typeinfo.info.type;
				if (!varCmd.typeinfo.IsTypeByRef() && type->specs[TYPE::SPEC_COPY_CONSTR]) {
					PreProcessProcPrintHeader(type->specs[TYPE::SPEC_COPY_CONSTR], globalInfo);
				}
			}
		}
	}

	for (int i=0; i < (int)type->subautos.size(); ++i) {
		SUBAUTO_INFO & si = type->subautos[i];

		if (si.relaxInfo.context.empty()) {
			continue;
		}

		const EVAL_CODE & code = si.proc->code;

		for (map<int,int>::const_iterator iter = si.relaxInfo.context.begin(); iter != si.relaxInfo.context.end(); ++iter) {
			const EVAL_COMMAND & labelCmd = code[iter->first];

			for (int i=0; i<(int)labelCmd.availVars.size(); ++i) {
				const EVAL_COMMAND & varCmd = code[labelCmd.availVars[i]];

				TYPE * type = varCmd.typeinfo.info.type;
				if (!varCmd.typeinfo.IsTypeByRef() && type->specs[TYPE::SPEC_COPY_CONSTR]) {
					PreProcessProcPrintHeader(type->specs[TYPE::SPEC_COPY_CONSTR], globalInfo);
				}
			}
		}
	}

	// preprocess default assign operator
	METHOD * defOpAssign = type->specs[TYPE::SPEC_DEF_OP_ASSIGN];
	if (defOpAssign != NULL && defOpAssign->impl == NULL) {
		PreProcessCodeBlockBegin(info, globalInfo, info.defAssignOpInfo);

		info.fieldTmpVars.resize(type->fields.size());

		for (int i=0; i<(int)type->fields.size(); ++i) {
			TYPE::FIELD & field = type->fields[i];

			PreProcessFieldCopy(field.type, info.defAssignOpInfo, info.fieldTmpVars[i], info, globalInfo);
		}

		// preprocess variables in relax stack
		if (!type->relaxInfo.context.empty()) {
			assert(type->specs[TYPE::SPEC_AUTO_PROC] != NULL);
			const EVAL_CODE & code = type->specs[TYPE::SPEC_AUTO_PROC]->code;

			info.relaxCopyTmpVars.resize(type->relaxInfo.context.size() + 1);

			for (map<int,int>::const_iterator iter = type->relaxInfo.context.begin(); iter != type->relaxInfo.context.end(); ++iter) {
				const EVAL_COMMAND & labelCmd = code[iter->first];
				assert(labelCmd.type == EVAL_COMMAND::TYPE_LABEL);

				INT_ARRAY & tmpVars = info.relaxCopyTmpVars[iter->second];
				tmpVars.resize(labelCmd.availVars.size());

				for (int i=0; i<(int)labelCmd.availVars.size(); ++i) {
					const EVAL_COMMAND & varCmd = code[labelCmd.availVars[i]];
					assert(varCmd.type == EVAL_COMMAND::TYPE_DECL_RELAX);

					PreProcessFieldCopy(varCmd.typeinfo, info.defAssignOpInfo, tmpVars[i], info, globalInfo);
				}
			}
		}

		for (int i=0; i < (int)type->subautos.size(); ++i) {
			SUBAUTO_INFO & si = type->subautos[i];

			if (si.relaxInfo.context.empty()) {
				continue;
			}

			const EVAL_CODE & code = si.proc->code;
			info.relaxCopyTmpVars.resize(type->relaxInfo.context.size() + 1);

			for (map<int,int>::const_iterator iter = si.relaxInfo.context.begin(); iter != si.relaxInfo.context.end(); ++iter) {
				const EVAL_COMMAND & labelCmd = code[iter->first];
				assert(labelCmd.type == EVAL_COMMAND::TYPE_LABEL);

				INT_ARRAY & tmpVars = info.relaxCopyTmpVars[iter->second];
				tmpVars.resize(labelCmd.availVars.size());

				for (int i=0; i<(int)labelCmd.availVars.size(); ++i) {
					const EVAL_COMMAND & varCmd = code[labelCmd.availVars[i]];
					assert(varCmd.type == EVAL_COMMAND::TYPE_DECL_RELAX);

					PreProcessFieldCopy(varCmd.typeinfo, info.defAssignOpInfo, tmpVars[i], info, globalInfo);
				}
			}
		}

		PreProcessCodeBlockEnd(info, globalInfo, info.defAssignOpInfo);
	}
}

/*************************************************************************/
/* TARGET_C::PrintRelaxCopy()                                            */
/*                                                                       */
/*************************************************************************/
void TARGET_C::PrintRelaxCopy(PRINT_RELAX_COPY_TYPE copyType, RELAX_INFO & relaxInfo, METHOD * method, INFO & localInfo, TYPE * type, const std::string & right)
{
	const bool isMainAuto = (method == type->specs[TYPE::SPEC_AUTO_PROC]);

	string thisVal = TAVR_THIS "->";
	if (type->IsTypeByRef()) {
		thisVal += "ptr->";
	}

	SUBAUTO_INFO * subautoInfo = NULL;
	if (!isMainAuto) {
		for (int i=0; i<(int)type->subautos.size(); ++i) {
			if (method == type->subautos[i].proc) {
				subautoInfo = &type->subautos[i];
				break;
			}
		}
		assert(subautoInfo != NULL);
	}

	bool needRight = true;
	if (copyType != PRCT_DESTR) {
		if (isMainAuto) {
			output << tabs << thisVal << TAVR_RELAX_CONTEXT " = ";
		} else {
			output << tabs << thisVal << TAVR_RELAX_CONTEXT "_" << subautoInfo->id << " = ";
		}

		if (right.empty()) {
			output << TAVR_SRC;
		} else {
			output << right;
		}

		if (isMainAuto) {
			output << "->" TAVR_RELAX_CONTEXT;
		} else {
			output << "->" TAVR_RELAX_CONTEXT "_" << subautoInfo->id;
		}
		output << " & ~" TAVR_RELAX_BUSY_BIT ";" << endl;
	} else {
		needRight = false;
	}

	assert(type->specs[TYPE::SPEC_AUTO_PROC] != NULL);
	const EVAL_CODE & code = method->code;

	if (isMainAuto) {
		output << tabs << "switch (" << thisVal << TAVR_RELAX_CONTEXT;
	} else {
		output << tabs << "switch (" << thisVal << TAVR_RELAX_CONTEXT "_" << subautoInfo->id;
	}
	output << ") {" << endl;
	AddTab();
	for (map<int,int>::const_iterator relaxIter = relaxInfo.context.begin(); relaxIter != relaxInfo.context.end(); ++relaxIter) {
		output << tabs << "case " << relaxIter->second << ":" << endl;
		AddTab();

		const EVAL_COMMAND & labelCmd = code[relaxIter->first];
		assert(labelCmd.type == EVAL_COMMAND::TYPE_LABEL);

		int offset = 0;

		typedef pair<string, string> STRING_PAIR;
		typedef map<int, STRING_PAIR> INT_STRING_PAIR_MAP;
		INT_STRING_PAIR_MAP names;

		for (int i=0; i < (int)labelCmd.availVars.size(); ++i) {
			const EVAL_COMMAND & varCmd = code[labelCmd.availVars[i]];
			assert(varCmd.type == EVAL_COMMAND::TYPE_DECL_RELAX);

			pair<INT_STRING_PAIR_MAP::iterator, bool> iterPair = names.insert(make_pair(labelCmd.availVars[i], STRING_PAIR()));
			INT_STRING_PAIR_MAP::iterator iter = iterPair.first;

			if (iterPair.second) {
				string & l = iter->second.first;
				l += "(*(";
				l += GetTypeName(varCmd.typeinfo);
				l += "*)(";
				l += thisVal;
				l += TAVR_RELAX_STACK;
				if (subautoInfo != NULL) {
					l += "_";
					l += ToString(subautoInfo->id);
				}

				l += "+";
				l += ToString(offset);
				l += "))";

				if (needRight) {
					string & r = iter->second.second;
					r += "(*(";
					r += GetTypeName(varCmd.typeinfo);
					r += "*)(";
					if (right.empty()) {
						r += TAVR_SRC;
					} else {
						r += right;
					}
					r += "->" TAVR_RELAX_STACK;
					if (subautoInfo != NULL) {
						r += "_";
						r += ToString(subautoInfo->id);
					}
					r += "+";
					r += ToString(offset);
					r += "))";
				}
			}

#ifdef _DEBUG
			output << tabs << "// " << project->Type2Str(varCmd.typeinfo) << " " << varCmd.info.name.str() << endl;
#endif
			switch (copyType) {
				case PRCT_COPY_CONSTR:
					PrintCopyConstr(varCmd.typeinfo, iter->second.first, iter->second.second);
					break;

				case PRCT_OP_EQ:
					localInfo.defAssignOpInfo.tmpVars[0] = localInfo.relaxCopyTmpVars[relaxIter->second][i];
					PrintAssignOp(false, varCmd.typeinfo, iter->second.first, localInfo.defAssignOpInfo, localInfo, NULL, NULL, &iter->second.second);
					break;

				case PRCT_DESTR:
					PrintDestructor(varCmd.typeinfo, iter->second.first);
					break;

				default:
					assert(!"Unknown copy type");
					break;
			}

			int size = TYPE_BY_REF_SIZE;
			if (!varCmd.typeinfo.IsTypeByRef()) {
				size = varCmd.typeinfo.info.type->size;
				assert(size > 0);
			}

			offset += size;
		}

		output << tabs << "break;" << endl;
		RemoveTab();
	}
	output << tabs << "default: if (" << thisVal << TAVR_RELAX_CONTEXT;
	if (!isMainAuto) {
		output << "_" << subautoInfo->id;
	}
	output << " != 0) " TAVR_UNKNOWN_RELAX_CONTEXT "(); break;" << endl;

	RemoveTab();
	output << tabs << "}" << endl;
}

/************************************************************************/
/* tavr::TARGET_C::PrintRelaxCopy
/*
/************************************************************************/
void TARGET_C::PrintRelaxCopy(PRINT_RELAX_COPY_TYPE copyType, INFO & localInfo, TYPE * type, const string & right)
{
	if (!type->relaxInfo.context.empty()) {
		assert(type->specs[TYPE::SPEC_AUTO_PROC] != NULL);
		PrintRelaxCopy(copyType, type->relaxInfo, type->specs[TYPE::SPEC_AUTO_PROC], localInfo, type, right);
	} else {
		assert(type->relaxInfo.stackSize == 0);
	}

	for (int i=0; i<(int)type->subautos.size(); ++i) {
		SUBAUTO_INFO & si = type->subautos[i];

		if (!si.relaxInfo.context.empty()) {
			PrintRelaxCopy(copyType, si.relaxInfo, si.proc, localInfo, type, right);
		} else {
			assert(si.relaxInfo.stackSize == 0);
		}
	}
}

/************************************************************************/
/* tavr::TARGET_C::ProcessType
/*
/************************************************************************/
void TARGET_C::ProcessType(TYPE * type, INFO & info, GLOBAL_INFO & globalInfo)
{
	assert(!type->methods.empty());

	// do not generate anything for extern auto
	if ((type->subtype == TYPE::SUBTYPE_AUTO || type->subtype == TYPE::SUBTYPE_EVENT) && type->attribs.is(ATTRIB_EXTERN)) {
		return ;
	}

	METHOD * defCopyConstr	= NULL;
	METHOD * defEventConstr = NULL;

	// generate type constructors
	for (int index = (int)type->methods.size() - 1; index >= 0; --index) {
		METHOD * method = type->methods[index];
		if (method->type != METHOD::TYPE_CONSTRUCTOR || method->attribs.is(ATTRIB_EXTERN)) {
			continue;
		}

		// check whether it is a default constructor
		if (method->impl == NULL) {
			if (method == type->specs[TYPE::SPEC_COPY_CONSTR]) {
				assert(defCopyConstr == NULL);
				defCopyConstr = method;
				continue;
			}
			if (method == type->specs[TYPE::SPEC_EVENT_CONSTR]) {
				assert(defEventConstr == NULL);
				defEventConstr = method;
				continue;
			}
		}

		PrintProcHeader(method, globalInfo);

		output << endl;

		ProcessCodeBlockBegin(info, info.constrCmdInfo);

		const char * thisVal = type->IsTypeByValue()
			? TAVR_THIS "->" : TAVR_THIS "->ptr->";

		if (method != type->specs[TYPE::SPEC_COPY_CONSTR]) {
			if (!type->relaxInfo.context.empty()) {
				output << "	" << thisVal << TAVR_RELAX_CONTEXT << " = 0;" << endl;
			}
			for (int i=0; i<(int)type->subautos.size(); ++i) {
				if (!type->subautos[i].relaxInfo.context.empty()) {
					output << "	" << thisVal << TAVR_RELAX_CONTEXT "_" << type->subautos[i].id << " = 0;" << endl;
				}
			}
		} else {
			assert(method->args.size() == 1);
			string right;
			if (method->args[0].name != project->emptyString) {
				right = method->args[0].name.str();
			} else {
				right = TAVR_DUMMY_ARG "0";
			}
			assert(method->args[0].type.isRef);

			PrintRelaxCopy(PRCT_COPY_CONSTR, info, type, right);
		}

		// initialize out subscriber list
		for (TYPE::OUT_TABLE::const_iterator iter = type->outTable.begin()
			, iterEnd = type->outTable.end()
			; iter != iterEnd
			; ++iter)
		{
			output << "	" << thisVal << TAVR_OUT_PREFIX << iter->first.str() << " = 0;" << endl;
		}

		for (int i=(int)type->fields.size() - 1; i >= 0; --i) {
			TYPE::FIELD & field = type->fields[i];

			INFO & fieldInfo = info.fieldInfo[i];

			if (field.value != NULL) {
				swap(info.argSets,		fieldInfo.argSets);
				swap(info.tmpVarSets,	fieldInfo.tmpVarSets);

				// hack tmp vars finalization, avoid variables result to be
				// destructed before new value was assigned
				PENDING_TMP_VAR_ARRAY pendingTmpVars;
				int lastCmpIndex = (int)type->fields[i].code.size() - 1;
				COMMAND_INFO & lastCmdInfo = fieldInfo.cmdInfo[lastCmpIndex];
				swap(pendingTmpVars,	lastCmdInfo.pendingTmpVars);

				PrintCode(field.code, fieldInfo);

				swap(pendingTmpVars,	lastCmdInfo.pendingTmpVars);
				swap(info.argSets,		fieldInfo.argSets);
				swap(info.tmpVarSets,	fieldInfo.tmpVarSets);

				assert(!fieldInfo.cmdInfo.empty());

				TYPEINFO & rightType = field.code.back().typeinfo;
				if (!rightType.IsNull()) {
					PrintCopyConstr(field.type, thisVal + field.name.str(), lastCmdInfo.value);
				} else {
					PrintVarDefInit(field.type, thisVal + field.name.str());
				}

				// finalize tmp vars
				assert(lastCmdInfo.argSet != -1);
				FinalizeTmpVars(lastCmdInfo.pendingTmpVars, info);
			} else {
				PrintVarDefInit(field.type, thisVal + field.name.str());
			}
		}

		info.constrInfo[index].curProc = method;
		PrintCode(method->code, info.constrInfo[index]);

		ProcessCodeBlockEnd(info, info.constrCmdInfo);
	}

	// generate copy constructor
	if (defCopyConstr != NULL) {
		info.curProc = defCopyConstr;
	
		PrintProcHeader(defCopyConstr, globalInfo);
		output << endl;

		assert(defCopyConstr->impl == NULL);

		// just use empty cmdInfo, there should not be any tmp variables
		ProcessCodeBlockBegin(info, info.copyConstrCmdInfo);

		string thisVal	= TAVR_THIS "->";
		string srcVal	= TAVR_SRC "->";
		assert(type->IsTypeByValue());

		// initialize out subscriber list
		for (TYPE::OUT_TABLE::const_iterator iter = type->outTable.begin()
			, iterEnd = type->outTable.end()
			; iter != iterEnd
			; ++iter)
		{
			output << "	" << thisVal << TAVR_OUT_PREFIX << iter->first.str() << " = 0;" << endl;
		}

		for (int i=0; i<(int)type->fields.size(); ++i) {
			const TYPE::FIELD & field = type->fields[i];

			PrintCopyConstr(field.type, thisVal + field.name.str(), srcVal + field.name.str());
		}

		// copy relax stack
		PrintRelaxCopy(PRCT_COPY_CONSTR, info, type);

		FinalizeProcArgs(defCopyConstr);

		ProcessCodeBlockEnd(info, info.copyConstrCmdInfo);
	}

	// generate copy constructor
	if (defEventConstr != NULL) {
		info.curProc = defEventConstr;
		
		PrintProcHeader(defEventConstr, globalInfo);
		output << endl;

		assert(defEventConstr->impl == NULL);

		// just use empty cmdInfo, there should not be any tmp variables
		ProcessCodeBlockBegin(info, info.eventConstrCmdInfo);

		const char * thisVal = TAVR_THIS "->ptr->";
		assert(type->IsTypeByRef());
		assert(type->outTable.empty());

		for (int i=0; i<(int)type->fields.size(); ++i) {
			const TYPE::FIELD & field = type->fields[i];
			PrintCopyConstr(field.type, thisVal + field.name.str(), field.name.str());
		}

		FinalizeProcArgs(defEventConstr);

		ProcessCodeBlockEnd(info, info.eventConstrCmdInfo);
	}

	// generate destructor
	METHOD * destr = type->specs[TYPE::SPEC_DESTR];
	if (destr != NULL && destr->impl == NULL) {
		PrintProcHeader(destr, globalInfo);
		output << endl << "{" << endl;
		AddTab();

		string thisVal = type->IsTypeByValue()
			? TAVR_THIS "->" : TAVR_THIS "->ptr->";

		// release subscribers lists
		for (TYPE::OUT_TABLE::const_iterator iter = type->outTable.begin()
			, iterEnd = type->outTable.end()
			; iter != iterEnd
			; ++iter)
		{
			output << "	" TAVR_RELEASE_OUT_LIST "(&" << thisVal << TAVR_OUT_PREFIX << iter->first.str() << ");" << endl;
		}

		// destruct fields
		for (int i=(int)type->fields.size() - 1; i >= 0; --i) {
			TYPE::FIELD & field = type->fields[i];
			PrintDestructor(field.type, thisVal + field.name.str());
		}

		// destruct relax stack
		PrintRelaxCopy(PRCT_DESTR, info, type);

		RemoveTab();
		output << "}" << endl;
	}

	// generate default assign operator if needed
	METHOD * defOpAssign = type->specs[TYPE::SPEC_DEF_OP_ASSIGN];
	if (defOpAssign != NULL && defOpAssign->impl == NULL) {
		PrintProcHeader(defOpAssign, globalInfo);
		output << endl;

		ProcessCodeBlockBegin(info, info.defAssignOpInfo);

		string thisVal	= TAVR_THIS "->";
		string srcVal	= TAVR_SRC "->";
		assert(type->IsTypeByValue());

		info.defAssignOpInfo.tmpVars.resize(1);

		for (int i=0; i<(int)type->fields.size(); ++i) {
			const TYPE::FIELD & field = type->fields[i];

			string right = srcVal;
			right += field.name.str();

			info.defAssignOpInfo.tmpVars[0] = info.fieldTmpVars[i];

			PrintAssignOp(false, field.type, thisVal + field.name.str(), info.defAssignOpInfo, info, NULL, NULL, &right);
		}

		PrintRelaxCopy(PRCT_OP_EQ, info, type);

		ProcessCodeBlockEnd(info, info.defAssignOpInfo);
	}
}

/************************************************************************/
/* tavr::TARGET_C::ProcessProc
/*
/************************************************************************/
void TARGET_C::ProcessProc(METHOD * proc, INFO & info, GLOBAL_INFO & globalInfo)
{
	const bool isAutoMethod = (proc->type == METHOD::TYPE_AUTO_PROC);

	// skip extern procs
	if (proc->type == METHOD::TYPE_CONSTRUCTOR || (proc->impl == NULL && !isAutoMethod)) {
		return ;
	}

	PrintProcHeader(proc, globalInfo);
	output << endl;

	const bool isAssignOpWRelax = proc->type == METHOD::TYPE_OPERATOR
		&&  proc->owner.type != NULL
		&&  proc->owner.type->specs[TYPE::SPEC_DEF_OP_ASSIGN] == proc
		&&	(!proc->owner.type->relaxInfo.context.empty()
		||  !proc->owner.type->subautos.empty());
	
	if (isAssignOpWRelax) {
		ProcessCodeBlockBegin(info, info.defAssignOpInfo);

		info.defAssignOpInfo.tmpVars.resize(1);

		assert(proc->args.size() == 1);
		string right;
		if (proc->args[0].name != project->emptyString) {
			right = proc->args[0].name.str();
		} else {
			right = TAVR_DUMMY_ARG "0";
		}
		if (!proc->args[0].type.isRef) {
			right = "(&" + right + ")";
		}
		PrintRelaxCopy(PRCT_OP_EQ, info, proc->owner.type, right);
	}

	// generate relax context restore
	string relaxContextVal;
	RELAX_INFO * relaxInfo = NULL;
	SUBAUTO_INFO * subautoInfo = NULL;

	bool hasBody = false;

	if (isAutoMethod) {
		TYPE * ownerType = proc->owner.type;

		bool isMainAuto = false;
		if (proc == ownerType->specs[TYPE::SPEC_AUTO_PROC]) {
			relaxInfo = &ownerType->relaxInfo;
			isMainAuto = true;
		} else {
			for (int i=0; i<(int)ownerType->subautos.size(); ++i) {
				SUBAUTO_INFO & si = ownerType->subautos[i];

				if (proc == si.proc) {
					subautoInfo = &si;
					relaxInfo   = &si.relaxInfo;
					break;
				}
			}
			assert(subautoInfo != NULL);
		}

		hasBody = !relaxInfo->context.empty() || isMainAuto && !ownerType->subautos.empty();
		if (hasBody) {
			output << "{" << endl;
			AddTab();

			// generate message translation for main auto proc
			if (isMainAuto && !ownerType->subautos.empty()) {
				output << tabs << "if (" TAVR_CUR_EVENT ".ref != 0) {" << endl;
				AddTab();
				output << tabs << TAVR_NULL_TYPE " " TAVR_TMP_EVENT " = " TAVR_CUR_EVENT ";" << endl;

				for (int i=0; i<(int)ownerType->subautos.size(); ++i) {
					SUBAUTO_INFO & si = ownerType->subautos[i];
					
					output << tabs << TAVR_ADD_POST_LAYER "();" << endl;
					if (i > 0) {
						output << tabs << TAVR_CUR_EVENT " = " TAVR_TMP_EVENT ";" << endl;
					}
					output << tabs << TAVR_INC_REF "(" TAVR_CUR_EVENT ");" << endl;
					output << tabs << GetProcName(si.proc) << "(" TAVR_THIS ");" << endl;
					output << tabs << TAVR_DEC_REF "(" TAVR_CUR_EVENT ");" << endl;
					output << tabs << TAVR_CUR_EVENT ".ptr = " TAVR_CUR_EVENT ".ref = 0;" << endl;
					output << tabs << TAVR_REMOVE_POST_LAYER "();" << endl;
				}

				output << tabs << TAVR_CUR_EVENT " = " TAVR_TMP_EVENT ";" << endl;
				RemoveTab();
				output << tabs << "} else {" << endl;
				AddTab();
				for (int i=0; i<(int)ownerType->subautos.size(); ++i) {
					SUBAUTO_INFO & si = ownerType->subautos[i];
					output << tabs << GetProcName(si.proc) << "(" TAVR_THIS ");" << endl;
				}
				RemoveTab();
				output << tabs << "}" << endl;
			}

			if (!relaxInfo->context.empty()) {
				if (ownerType->IsTypeByRef()) {
					relaxContextVal += TAVR_THIS "->ptr->" TAVR_RELAX_CONTEXT;
				} else {
					relaxContextVal += TAVR_THIS "->" TAVR_RELAX_CONTEXT;
				}
				if (subautoInfo != NULL) {
					relaxContextVal += "_";
					relaxContextVal += ToString(subautoInfo->id);
				}

				output << "	switch (" << relaxContextVal << ") {" << endl;

				if (ownerType->IsTypeByValue()) {
					if (subautoInfo == NULL) {
						output << "		" TAVR_RELAX_CASE_DEFAULT "();" << endl;
					} else {
						output << "		" TAVR_RELAX_CASE_DEFAULT_S "(" << subautoInfo->id << ");" << endl;
					}
				} else {
					if (subautoInfo == NULL) {
						output << "		" TAVR_RELAX_CASE_DEFAULT2 "();" << endl;
					} else {
						output << "		" TAVR_RELAX_CASE_DEFAULT2_S "(" << subautoInfo->id << ");" << endl;
					}
				}

				for (std::map<int, int>::const_iterator iter = relaxInfo->context.begin()
					, iterEnd = relaxInfo->context.end()
					; iter != iterEnd
					; ++iter)
				{
					assert(proc->code[iter->first].type == EVAL_COMMAND::TYPE_LABEL);
					if (ownerType->IsTypeByValue()) {
						if (subautoInfo == NULL) {
							output << "		" TAVR_RELAX_CASE "(" << iter->second << ", "
								<< info.cmdInfo[iter->first].value << ");" << endl;
						} else {
							output << "		" TAVR_RELAX_CASE_S "(" << subautoInfo->id << ", " << iter->second << ", "
								<< info.cmdInfo[iter->first].value << ");" << endl;
						}
					} else {
						if (subautoInfo == NULL) {
							output << "		" TAVR_RELAX_CASE2 "(" << iter->second << ", "
								<< info.cmdInfo[iter->first].value << ");" << endl;
						} else {
							output << "		" TAVR_RELAX_CASE2_S "(" << subautoInfo->id << ", " << iter->second << ", "
								<< info.cmdInfo[iter->first].value << ");" << endl;
						}
					}
				}

				output << "		default: " TAVR_UNKNOWN_RELAX_CONTEXT "(); return ;" << endl;
				output << "	}" << endl;
			}
		}
	}
	
	info.curProc = proc;
	PrintCode(proc->code, info);

	// after all it should be default context restoration
	if (hasBody) {
		assert(isAutoMethod);

		assert(relaxInfo != NULL);
		RemoveTab();
		if (!relaxInfo->context.empty()) {
			output << "	" << relaxContextVal << " = 0;" << endl;
		}
		output << "}" << endl;
	} else if (isAssignOpWRelax) {
		ProcessCodeBlockEnd(info, info.defAssignOpInfo);
	}
}

/************************************************************************/
/* tavr::TARGET_C::ProcessCodeBlockBegin
/*
/************************************************************************/
void TARGET_C::ProcessCodeBlockBegin(INFO & localInfo, COMMAND_INFO & cmdInfo, const std::vector<EVAL_COMMAND> * code, int index)
{
	++localInfo.blocksCounter;

	output << tabs << "{" << endl;
	AddTab();

	assert(cmdInfo.argSet != -1);

	if (code != NULL) {
		INT_ARRAY & argSet = localInfo.argSets[cmdInfo.argSet];
		for (int k=0; k<(int)argSet.size(); ++k) {
			const EVAL_COMMAND & command = (*code)[ argSet[k] ];
			
			if (command.type == EVAL_COMMAND::TYPE_DECL) {
				assert(command.typeinfo.type == TYPEINFO::TYPE_TYPE);
				output << tabs << GetTypeName(command.typeinfo) << " " << command.info.name.c_str() << ";" << endl;
			} else {
				assert(command.type == EVAL_COMMAND::TYPE_DECL_RELAX);
			}
		}
	} else {
		assert(localInfo.argSets[cmdInfo.argSet].empty());
	}

	TMP_VAR_ARRAY & tmpVarSet = localInfo.tmpVarSets[cmdInfo.argSet];
	for (int k=0; k<(int)tmpVarSet.size(); ++k) {
		output << tabs << GetTypeName(tmpVarSet[k].type) << " " TAVR_TMP << tmpVarSet[k].index << ";" << endl;
	}
}

/*************************************************************************/
/* TARGET_C::PrintArrDestrBody()                                         */
/*                                                                       */
/*************************************************************************/
void TARGET_C::PrintArrDestrBody(const TYPEINFO & ti, const std::string & value)
{
	assert(ti.array > 0);

	METHOD * destr = ti.info.type->specs[TYPE::SPEC_DESTR];

	TYPEINFO elemType = ti;
	--elemType.array;

	if (elemType.IsTypeByRef()) {
		output << tabs << TAVR_DESTR_ARRAY_REF "(" << value << ");" << endl;
		return ;
	}

	output << tabs << TAVR_DESTR_ARRAY_VALUE "(" << value << ", " << GetProcName(destr) << ");" << endl;
}

/************************************************************************/
/* tavr::TARGET_C::PrintDestructor
/*
/************************************************************************/
void TARGET_C::PrintDestructor(const TYPEINFO & ti, const std::string & value)
{
	if (ti.IsTypeByRef()) {
		output << tabs << TAVR_DEC_REF "(" << value << ");" << endl;
		return ;
	}

	METHOD * destr = ti.info.type->specs[TYPE::SPEC_DESTR];
	if (destr == NULL) {
		return ;
	}

	output << tabs << GetProcName(destr) << "(&" << value << ");" << endl;
}

/************************************************************************/
/* tavr::TARGET_C::FinalizeLocalVars
/*
/************************************************************************/
void TARGET_C::FinalizeLocalVars(COMMAND_INFO & cmdInfo, INFO & localInfo, const EVAL_CODE & code, int returnCmdIndex /*= -1*/)
{
	INT_ARRAY arr;
	swap(arr, localInfo.argSets[cmdInfo.argSet]);
	for (int i=(int)arr.size() - 1; i >= 0; --i) {
		int index = arr[i];
		if (returnCmdIndex != -1 && index > returnCmdIndex) {
			continue;
		}

		const EVAL_COMMAND & cmd = code[index];
		assert(cmd.type == EVAL_COMMAND::TYPE_DECL
			|| cmd.type == EVAL_COMMAND::TYPE_DECL_RELAX);

#ifdef _DEBUG
		output << tabs << "// destr: " << cmd.info.name.str() << endl;
#endif

		if (cmd.type != EVAL_COMMAND::TYPE_DECL_RELAX) {
			PrintDestructor(cmd.typeinfo, cmd.info.name.str());
		} else {
			PrintDestructor(cmd.typeinfo, localInfo.cmdInfo[index].value);
		}
	}
}

/*************************************************************************/
/* TARGET_C::FinalizeLocalVarsReturn()                                   */
/*                                                                       */
/*************************************************************************/
void TARGET_C::FinalizeLocalVarsReturn(COMMAND_INFO & cmdInfo, INFO & localInfo, const EVAL_CODE & code, int returnCmdIndex)
{
	FinalizeLocalVars(cmdInfo, localInfo, code, returnCmdIndex);

	for (int i=0; i<(int)localInfo.argSets.size(); ++i) {
		INT_ARRAY & arr = localInfo.argSets[i];

		for (int i=(int)arr.size() - 1; i >= 0; --i) {
			int index = arr[i];
			if (returnCmdIndex != -1 && index > returnCmdIndex) {
				continue;
			}

			const EVAL_COMMAND & cmd = code[index];
			assert(cmd.type == EVAL_COMMAND::TYPE_DECL
				|| cmd.type == EVAL_COMMAND::TYPE_DECL_RELAX);

#ifdef _DEBUG
			output << tabs << "// destr: " << cmd.info.name.str() << endl;
#endif

			if (cmd.type != EVAL_COMMAND::TYPE_DECL_RELAX) {
				PrintDestructor(cmd.typeinfo, cmd.info.name.str());
			} else {
				PrintDestructor(cmd.typeinfo, localInfo.cmdInfo[index].value);
			}
		}
	}
}

/************************************************************************/
/* tavr::TARGET_C::FinalizeProcArgs
/*
/************************************************************************/
void TARGET_C::FinalizeProcArgs(METHOD * proc)
{
	string argName;
	for (int i=(int)proc->args.size() - 1; i >= 0; --i) {
		METHOD::ARG & arg = proc->args[i];

		// do not call destructors for reference args
		if (arg.type.isRef) {
			continue;
		}

		if (arg.name == project->emptyString) {
			argName = TAVR_DUMMY_ARG + ToString(i);
		} else {
			argName = arg.name.str();
		}

#ifdef _DEBUG
		output << tabs << "// destr: " << argName << endl;
#endif

		PrintDestructor(arg.type, argName);
	}
}

/************************************************************************/
/* tavr::TARGET_C::FinalizeGotoVars
/*
/************************************************************************/
void TARGET_C::FinalizeGotoVars(const std::vector<int> & availVars, const EVAL_CODE & code, INFO & localInfo)
{
	for (int index = (int)availVars.size() - 1; index >= 0; --index) {
		const EVAL_COMMAND & varDeclCmd = code[availVars[index]];
		
#ifdef _DEBUG
		output << tabs << "// destr: " << varDeclCmd.info.name.str() << endl;
#endif

		if (varDeclCmd.type != EVAL_COMMAND::TYPE_DECL_RELAX) {
			PrintDestructor(varDeclCmd.typeinfo, varDeclCmd.info.name.str());
		} else {
			PrintDestructor(varDeclCmd.typeinfo, localInfo.cmdInfo[availVars[index]].value);
		}
	}
}

/************************************************************************/
/* tavr::TARGET_C::ProcessCodeBlockEnd
/*
/************************************************************************/
void TARGET_C::ProcessCodeBlockEnd(INFO & localInfo, COMMAND_INFO & cmdInfo, const EVAL_CODE * code, int index)
{
	--localInfo.blocksCounter;
	assert(localInfo.blocksCounter >= 0);

	// call destructors for all local variables
	if (code != NULL) {
		FinalizeLocalVars(cmdInfo, localInfo, *code);

		if (localInfo.blocksCounter == 0) {
			assert(localInfo.curProc != NULL);
			FinalizeProcArgs(localInfo.curProc);
		}
	}

	RemoveTab();
	output << tabs << "}" << endl;
}

/************************************************************************/
/* tavr::TARGET_C::FinalizeTmpVars
/*
/************************************************************************/
void TARGET_C::FinalizeTmpVars(PENDING_TMP_VAR_ARRAY & pendingTmpVars, INFO & localInfo)
{
	if (pendingTmpVars.empty()) {
		return ;
	}

	PENDING_TMP_VAR_ARRAY tmpVars;
	swap(tmpVars, pendingTmpVars);

	for (int i=(int)tmpVars.size() - 1; i >= 0; --i) {
#ifdef _DEBUG
		output << tabs << "// destr: " TAVR_TMP << tmpVars[i].index;
		if (!tmpVars[i].needDestructor) {
			output << " : not needed";
		}
		output << endl;
#endif

		if (!tmpVars[i].needDestructor) {
			continue;
		}

		TMP_VAR_ARRAY & curTmpVars = localInfo.tmpVarSets[tmpVars[i].argSet];

		bool found = false;
		for (int k=(int)curTmpVars.size() - 1; k >= 0; --k) {
			TMP_VAR & var = curTmpVars[k];
			if (var.index != tmpVars[i].index) {
				continue;
			}

			PrintDestructor(var.type, TAVR_TMP + ToString(var.index));

			found = true;
			break;
		}
	}
}

/************************************************************************/
/* tavr::TARGET_C::PrintProcArgs
/*
/************************************************************************/
std::string TARGET_C::ProcessProcArgs(bool thisCall, METHOD * proc, COMMAND_INFO & cmdInfo, INFO & info, const EVAL_COMMAND * cmd, const EVAL_CODE * code, std::string * right /*= NULL*/)
{
	const int numArgs = (int)proc->args.size();

	assert((int)cmdInfo.tmpVars.size() >= numArgs);

	std::string res;
	res.reserve(128);

	if (cmd != NULL) {
		for (int i=0; i<numArgs; ++i) {
			TYPEINFO & argType = proc->args[i].type;

			if (argType.isRef) {
				res += "&";
				res += info.cmdInfo[cmd->args[i + thisCall]].value;
			} else {
				assert(cmdInfo.tmpVars[i] != -1);

				string tmpVal = TAVR_TMP + ToString(cmdInfo.tmpVars[i]);
				const TYPEINFO & rightType = (*code)[cmd->args[i + thisCall]].typeinfo;
				if (!rightType.IsNull()) {
					PrintCopyConstr(argType, tmpVal, info.cmdInfo[cmd->args[i + thisCall]].value);
				} else {
					PrintVarDefInit(argType, tmpVal);
				}

				res += tmpVal;
			}

			if (i + 1 < numArgs) {
				res += ", ";
			}
		}

		return res;
	}

	// support some hack
	assert(right != NULL && proc->args.size() == 1);

	TYPEINFO & argType = proc->args[0].type;
	if (argType.isRef) {
		res += "&";
		res += *right;
	} else {
		assert(cmdInfo.tmpVars[0] != -1);

		string tmpVal = TAVR_TMP + ToString(cmdInfo.tmpVars[0]);
		PrintCopyConstr(argType, tmpVal, *right);
		res += tmpVal;
	}

	return res;
}

/************************************************************************/
/* tavr::TARGET_C::PrintCode
/*
/************************************************************************/
void TARGET_C::PrintCode(const std::vector<EVAL_COMMAND> & code, INFO & localInfo)
{
	std::vector<COMMAND_INFO> & info = localInfo.cmdInfo;

	for (int i=0; i<(int)code.size(); ++i) {
		const EVAL_COMMAND & cmd = code[i];
		COMMAND_INFO & cmdInfo = info[i];

		switch (cmd.type) {
			case EVAL_COMMAND::TYPE_CONSTANT:
				assert(cmd.typeinfo.type == TYPEINFO::TYPE_R_VALUE);
				assert(!cmd.typeinfo.isRef);
				if (cmd.typeinfo.IsIntegral() || cmd.typeinfo.IsFloating()) {
					cmdInfo.value = cmd.info.name.str();
				} else 
				if (cmd.typeinfo.IsBoolean()) {
					cmdInfo.value = (cmd.info.name.str() == "true") ? "1" : "0";
				} else
				if (cmd.typeinfo.info.type->subtype == TYPE::SUBTYPE_NULL) {
					assert(cmd.info.name.str() == "null");
				} else
				if (cmd.typeinfo.info.type->subtype == TYPE::SUBTYPE_CHARACTER) {
					cmdInfo.value = "L" + cmd.info.name.str();
				} else
				if (cmd.typeinfo.info.type == project->defType[PROJECT::DEF_TYPE_STRING]) {
#ifdef _DEBUG
					output << tabs << "// " << cmd.info.name.str() << endl;
#endif
					assert(cmdInfo.tmpVars.size() == 1);
					string tmpName = TAVR_TMP + ToString(cmdInfo.tmpVars[0]);

					output << tabs << tmpName << ".ref = " TAVR_ALLOC_STRING "(L"
						<< cmd.info.name.str() << ");" << endl;
					output << tabs << tmpName << ".ptr = (void*)(" << tmpName << ".ref+" TAVR_OFFSET ");" << endl;

					cmdInfo.value = tmpName;
				} else {
					assert(!"Unknown output typeinfo");
				}
				break;

			case EVAL_COMMAND::TYPE_VARIABLE:
				assert(cmd.args.size() == 1);
				if (cmd.info.proc == NULL) {
					assert(code[cmd.args[0]].type == EVAL_COMMAND::TYPE_DECL
						|| code[cmd.args[0]].type == EVAL_COMMAND::TYPE_DECL_RELAX);

					assert(code[cmd.args[0]].info.name == cmd.info.name);

					const EVAL_COMMAND & relaxVarCmd = code[cmd.args[0]];
					if (relaxVarCmd.type == EVAL_COMMAND::TYPE_DECL_RELAX) {
						cmdInfo.value = info[cmd.args[0]].value;
						break;
					}
				} else {
					assert(cmd.info.proc->args[cmd.args[0]].name == cmd.info.name);
				}

				if (!cmd.typeinfo.isRef) {
					cmdInfo.value = cmd.info.name.str();
				} else {
					cmdInfo.value = "(*";
					cmdInfo.value += cmd.info.name.str();
					cmdInfo.value += ")";
				}
				break;

			case EVAL_COMMAND::TYPE_THIS:
				if (cmd.isUsed) {
					cmdInfo.value = "(*" TAVR_THIS ")";
				}
				break;

			case EVAL_COMMAND::TYPE_EVENT:
				if (cmd.isUsed) {
					cmdInfo.value = "(*(";
					cmdInfo.value += GetTypeName(cmd.typeinfo);
					cmdInfo.value += "*)&" TAVR_CUR_EVENT ")";
				}
				break;

			case EVAL_COMMAND::TYPE_CODE_BLOCK_BEGIN:
				ProcessCodeBlockBegin(localInfo, cmdInfo, &code, i);
				break;

			case EVAL_COMMAND::TYPE_CODE_BLOCK_END:
				ProcessCodeBlockEnd(localInfo, cmdInfo, &code, i);
				break;

			case EVAL_COMMAND::TYPE_DECL:
			case EVAL_COMMAND::TYPE_DECL_RELAX:
				{
					const bool isRelax = (cmd.type == EVAL_COMMAND::TYPE_DECL_RELAX);

#ifdef _DEBUG
					output << tabs << "// ";
					if (isRelax) {
						output << "relax : ";
					}
					output << project->Type2Str(cmd.typeinfo) << " "
						<< cmd.info.name.str();
					if (cmd.info.proc != NULL) {
						output << " = " << project->Type2Str(cmd.info.proc);
					} else if (!cmd.args.empty()) {
						output << " = " << project->Type2Str(code[cmd.args[0]].typeinfo);
					}
					output << endl;
#endif

					string value;
					if (!isRelax) {
						value = cmd.info.name.str();
					} else {
						value = cmdInfo.value;
					}

					assert(!cmd.typeinfo.isRef);

					if (cmd.info.proc == NULL) {
						if (cmd.args.empty()) {
							PrintVarDefInit(cmd.typeinfo, value);
						} else {
							assert(cmd.args.size() == 1);
							METHOD * copyConstr = cmd.typeinfo.info.type->specs[TYPE::SPEC_COPY_CONSTR];

							if (cmd.typeinfo.IsTypeByValue() && copyConstr != NULL) {
								output << tabs << GetProcName(copyConstr) << "(&" << value << ", ";
								if (copyConstr->args[0].type.isRef) {
									output << "&";
								}
								output << info[cmd.args[0]].value << ");" << endl;
							} else if (!cmd.typeinfo.IsTypeByRef()) {
								output << tabs << value << " = " << info[cmd.args[0]].value << ";" << endl;
							} else {
								const EVAL_COMMAND & rightCmd = code[cmd.args[0]];
								const TYPEINFO & rightType = rightCmd.typeinfo;
								if (!rightType.IsNull() || rightCmd.type != EVAL_COMMAND::TYPE_CONSTANT) {
									// it could be null type, but if it's not a constant, then it's a value! :)
									output << tabs << value << " = " << info[cmd.args[0]].value << ";" << endl;
									output << tabs << TAVR_INC_REF "(" << value << ");" << endl;
								} else {
									PrintVarDefInit(cmd.typeinfo, value);
								}
							}
						}
						break;
					} 

					assert(cmd.info.proc->IsMethod());
					assert(cmd.info.proc->type == METHOD::TYPE_CONSTRUCTOR);

					assert(cmd.args.size() == cmd.info.proc->args.size());

					string args = ProcessProcArgs(false, cmd.info.proc, cmdInfo, localInfo, &cmd, &code);

					output << tabs << GetProcName(cmd.info.proc) << "(&" << value;
					if (!cmd.args.empty()) {
						output << ", ";
					}
					output << args << ");" << endl;
				}
				break;

			case EVAL_COMMAND::TYPE_RETURN:
				{
					if (cmd.args.empty()) {
						FinalizeLocalVarsReturn(cmdInfo, localInfo, code, i);
						assert(localInfo.curProc != NULL);
						FinalizeProcArgs(localInfo.curProc);

						output << tabs << "return ;" << endl;
						break;
					}

					int index = cmd.args[0];

					bool needTmp = (cmd.typeinfo.IsTypeByRef()
						|| cmd.typeinfo.info.type->specs[TYPE::SPEC_COPY_CONSTR] != NULL);

					if (needTmp) {
						assert(cmdInfo.tmpVars.size() == 1);

						string resTmp = TAVR_TMP + ToString(cmdInfo.tmpVars[0]);

						const TYPEINFO & rightType = code[index].typeinfo;
						if (!rightType.IsNull()) {
							PrintCopyConstr(cmd.typeinfo, resTmp, info[index].value);
						} else {
							PrintVarDefInit(cmd.typeinfo, resTmp);
						}

						assert(cmdInfo.argSet != -1);
						FinalizeTmpVars(cmdInfo.pendingTmpVars, localInfo);

						// TODO: need refactoring
						FinalizeLocalVarsReturn(cmdInfo, localInfo, code, i);
						assert(localInfo.curProc != NULL);
						FinalizeProcArgs(localInfo.curProc);

						output << tabs << "return " << resTmp << ";" << endl;
					} else {
						FinalizeLocalVarsReturn(cmdInfo, localInfo, code, i);
						assert(localInfo.curProc != NULL);
						FinalizeProcArgs(localInfo.curProc);

						output << tabs << "return " << info[index].value << ";" << endl;
					}
				}
				break;

			case EVAL_COMMAND::TYPE_RELAX:
				{
					assert(cmd.args.size() == 1);
					assert(localInfo.curProc != NULL);
					assert(localInfo.curProc->type == METHOD::TYPE_AUTO_PROC);
					TYPE * type = localInfo.curProc->owner.type;

					// call destructors for necessary variables
					FinalizeGotoVars(cmd.availVars, code, localInfo);

					// find relax info
					RELAX_INFO * relaxInfo = NULL;
					SUBAUTO_INFO * subautoInfo = NULL;
					if (localInfo.curProc == type->specs[TYPE::SPEC_AUTO_PROC]) {
						relaxInfo = &type->relaxInfo;
					} else {
						for (int i=0; i<(int)type->subautos.size(); ++i) {
							SUBAUTO_INFO & si = type->subautos[i];
							if (si.proc == localInfo.curProc) {
								relaxInfo = &si.relaxInfo;
								subautoInfo = &si;
								break;
							}
						}
						assert(subautoInfo != NULL);
					}
					assert(relaxInfo != NULL);

					// store relax context
					assert(relaxInfo->context.find(cmd.args[0]) != relaxInfo->context.end());
					output << tabs << TAVR_THIS "->";
					if (type->IsTypeByRef()) {
						output << "ptr->";
					}
					output << TAVR_RELAX_CONTEXT;
					if (subautoInfo != NULL) {
						output << "_" << subautoInfo->id;
					}
					output << " = " << relaxInfo->context[cmd.args[0]] << ";" << endl;
					output << tabs << "return ;" << endl;
				}
				break;

			case EVAL_COMMAND::TYPE_THIS_CALL:
			case EVAL_COMMAND::TYPE_PROC_CALL:
			case EVAL_COMMAND::TYPE_CONSTR_CALL:
				{
#ifdef _DEBUG
					output << tabs << "// " << project->Type2Str(cmd.info.proc) << endl;
#endif

					// member call
					const bool thisCall = (cmd.type == EVAL_COMMAND::TYPE_THIS_CALL);

					// constructor call
					const bool isConstr = (cmd.type == EVAL_COMMAND::TYPE_CONSTR_CALL);

					assert(cmd.args.size() - thisCall == cmd.info.proc->args.size());
					assert(!isConstr || cmd.typeinfo.IsTypeByValue());

					METHOD * proc = cmd.info.proc;

					// add layer on 'auto' method call
					const bool isAutoProc = proc->type == METHOD::TYPE_METHOD
						&& proc == proc->owner.type->specs[TYPE::SPEC_AUTO_PROC];

					if (isAutoProc) {
						output << tabs << TAVR_ADD_POST_LAYER "();" << endl;
					}

					// proc name
					string val;
					val.reserve(128);

					val += GetProcName(cmd.info.proc);
					val += "(";

					// first argument for this and constructor call
					if (thisCall) {
						val += "&";
						val += info[cmd.args[0]].value;
						if (cmd.args.size() > 1) {
							val += ", ";
						}
					} else if (isConstr) {
						assert(cmdInfo.tmpVars.size() == proc->args.size() + 1);
						val += "&" TAVR_TMP;
						val += ToString(cmdInfo.tmpVars.back());
						if (!cmd.args.empty()) {
							val += ", ";
						}
					}

					// procedure arguments
					val += ProcessProcArgs(thisCall, cmd.info.proc, cmdInfo, localInfo, &cmd, &code);
					val += ")";

					// proc result
					if (cmd.info.proc->result.IsTypeByValue() || cmd.info.proc->result.IsTypeByRef()) {
						assert(cmdInfo.tmpVars.size() == proc->args.size() + 1);

						// last tmp var is for result
						cmdInfo.value = TAVR_TMP + ToString(cmdInfo.tmpVars.back());
						output << tabs << info[i].value << " = " << val << ";" << endl;
					} else if (isConstr) {
						assert(cmdInfo.tmpVars.size() == proc->args.size() + 1);

						info[i].value = TAVR_TMP + ToString(cmdInfo.tmpVars.back());
						output << tabs << val << ";" << endl;
					} else {
						if (cmd.isUsed) {
							swap(cmdInfo.value, val);
						} else {
							output << tabs << val << ";" << endl;
						}
					}

					if (isAutoProc) {
						output << tabs << TAVR_REMOVE_POST_LAYER "();" << endl;
					}
				}
				break;

			case EVAL_COMMAND::TYPE_NEW_ARRAY:
				{
#ifdef _DEBUG
					{
						TYPEINFO ti;
						ti = cmd.typeinfo;
						assert(ti.array > 0);
						--ti.array;
						output << tabs << "// new " << project->Type2Str(ti) << "[]" << endl;
					}
#endif

					assert(cmd.args.size() == 1);

					assert(info[i].tmpVars.size() == 2);

					string tmpCount = TAVR_TMP + ToString(info[i].tmpVars[0]);
					string tmpRes   = TAVR_TMP + ToString(info[i].tmpVars[1]);

					output << tabs << tmpCount << " = " << info[cmd.args[0]].value << ";" << endl;

					TYPEINFO elemType = cmd.typeinfo;
					assert(elemType.array >= 0);
					--elemType.array;

					int sizeofElemVal = TYPE_BY_REF_SIZE;
					if (!elemType.IsTypeByRef()) {
						sizeofElemVal = elemType.info.type->size;
						assert(sizeofElemVal > 0);
					}
					string sizeofElem = ToString(sizeofElemVal);

					output << tabs << tmpRes << ".ref = "
						<< TAVR_ALLOC_ARRAY "(" << tmpCount << ", "
						<< sizeofElem << ", " TAVR_TYPEINFO << GetTypeName(cmd.typeinfo, true) << ");" << endl;
					output << tabs << tmpRes << ".ptr = (void*)(" << tmpRes << ".ref+" TAVR_ARRAY_OFFSET ");" << endl;

					if (elemType.IsTypeByValue()) {
						output << tabs << "while (--" << tmpCount << ">= 0) {" << endl;
						string arg = tmpRes; arg += ".ptr["; arg += tmpCount; arg += "]";
						PrintVarDefInit(elemType, arg);
						output << tabs << "}" << endl;
					} else {
						output << tabs << "memset(" << tmpRes << ".ptr, 0, "
							<< sizeofElem << "*" << tmpCount << ");" << endl;
					}

					swap(info[i].value, tmpRes);
				}
				break;

			case EVAL_COMMAND::TYPE_NEW_AUTO:
				{
#ifdef _DEBUG
					output << tabs << "// new " << project->Type2Str(cmd.info.proc) << endl;
#endif

					assert(cmd.typeinfo.info.type->size > 0);
					assert(info[i].tmpVars.size() == 1 + (int)cmd.info.proc->args.size());

					string tmpRes   = TAVR_TMP + ToString(info[i].tmpVars.back());
					string argsVal	= ProcessProcArgs(false, cmd.info.proc, cmdInfo, localInfo, &cmd, &code);

					assert(cmd.typeinfo.IsTypeByRef() && cmd.typeinfo.array == 0);
					const bool isExtern = cmd.typeinfo.info.type->attribs.is(ATTRIB_EXTERN);

					assert(cmd.info.proc->type == METHOD::TYPE_CONSTRUCTOR);

					if (!isExtern) {
						output << tabs << tmpRes << ".ref = "
							<< TAVR_ALLOC "("
							<< ToString(cmd.typeinfo.info.type->size)
							<< ", " TAVR_TYPEINFO << GetTypeName(cmd.typeinfo, true) << ");" << endl;
						output << tabs << tmpRes << ".ptr = (void*)(" << tmpRes << ".ref+" TAVR_OFFSET ");" << endl;

						output << tabs << GetProcName(cmd.info.proc);

						output << "(&" << tmpRes;
						if (!cmd.args.empty()) {
							output << ", ";
						}
						output << argsVal << ");" << endl;
					} else {
						output << tabs << tmpRes << ".ref = "
							<< GetProcName(cmd.info.proc) << "(" << argsVal << ");" << endl;

						output << tabs << tmpRes << ".ptr = (void*)(" << tmpRes << ".ref+" TAVR_OFFSET ");" << endl;
					}

					swap(info[i].value, tmpRes);
				}
				break;

			case EVAL_COMMAND::TYPE_ARRAY_LENGTH:
				{
#ifdef _DEBUG
					output << tabs << "// .length" << endl;
#endif
					assert(cmd.args.size() == 1);
					assert(cmdInfo.tmpVars.size() == 1);

					string tmp = TAVR_TMP + ToString(cmdInfo.tmpVars[0]);

					output << tabs << tmp << " = " << info[cmd.args[0]].value << ";" << endl;
					output << tabs << TAVR_ASSERT_NULL_PTR "(" << tmp << ");" << endl;

					tmp = TAVR_ARRAY_LENGTH "(" + tmp + ")";

					swap(cmdInfo.value, tmp);
				}
				break;

			case EVAL_COMMAND::TYPE_ARRAY_INDEX:
			case EVAL_COMMAND::TYPE_STRING_INDEX:
				{
#ifdef _DEBUG
					output << tabs << "// " << project->Type2Str(cmd.typeinfo) << "[index]" << endl;
#endif
					assert(cmd.args.size() == 2);
					assert(cmdInfo.tmpVars.size() == 2);

					const bool isArray = (cmd.type == EVAL_COMMAND::TYPE_ARRAY_INDEX);

					string tmp		= TAVR_TMP + ToString(cmdInfo.tmpVars[0]);
					string tmpIndex	= TAVR_TMP + ToString(cmdInfo.tmpVars[1]);

					output << tabs << tmp << " = " << info[cmd.args[0]].value << ";" << endl;
					output << tabs << tmpIndex << " = " << info[cmd.args[1]].value << ";" << endl;

					std::string val;

					if (isArray) {
						output << tabs << TAVR_ASSERT_ARRAY_INDEX "(" << tmp << ", " << tmpIndex << ");" << endl;

						val += tmp;
						val += ".ptr[";
						val += tmpIndex;
						val += "]";
					} else {
						output << tabs << TAVR_ASSERT_STRING_INDEX "(" << tmp << ", " << tmpIndex << ");" << endl;

						val += "((__tavr_char *)(";
						val += tmp;
						val += ".ref+" TAVR_ARRAY_OFFSET "))[";
						val += tmpIndex;
						val += "]";
					}

					swap(cmdInfo.value, val);
				}
				break;

			case EVAL_COMMAND::TYPE_IF:
				{
					assert(cmd.args.size() == 1);

					string condition;
					if (!cmdInfo.pendingTmpVars.empty()) {
						assert(cmdInfo.tmpVars.size() == 1);

						condition = TAVR_TMP + ToString(cmdInfo.tmpVars[0]);
						output << tabs << condition << " = " << info[cmd.args[0]].value << ";" << endl;

						assert(cmdInfo.argSet != -1);
						FinalizeTmpVars(cmdInfo.pendingTmpVars, localInfo);
					} else {
						assert(cmdInfo.tmpVars.empty());
						swap(condition, info[cmd.args[0]].value);
					}

					output << tabs << "if(" << condition << ")" << endl;
				}
				break;

			case EVAL_COMMAND::TYPE_ELSE:
				assert(cmd.args.empty());
				output << tabs << "else" << endl;
				break;

			case EVAL_COMMAND::TYPE_LABEL:
				if (cmd.isUsed) {
					output << cmdInfo.value << ": ;" << endl;
				}
				break;

			case EVAL_COMMAND::TYPE_GOTO:
				assert(cmd.args.size() == 1);
				assert(code[cmd.args[0]].type == EVAL_COMMAND::TYPE_LABEL);	
				
				FinalizeGotoVars(cmd.availVars, code, localInfo);

				output << tabs << "goto " << info[cmd.args[0]].value << ";" << endl;
				break;

			case EVAL_COMMAND::TYPE_SUBNAME:
				{
					assert(cmd.args.size() == 1);

#ifdef _DEBUG
					if (code[cmd.args[0]].typeinfo.IsTypeByRef()) {
						output << tabs << "// ." << cmd.info.name.str() << endl;
					}
#endif

					assert(cmd.typeinfo.IsValue());
					assert(code[cmd.args[0]].typeinfo.IsUserType());
					const int argIndex = cmd.args[0];
					string val = info[argIndex].value;
	
					if (code[argIndex].typeinfo.IsTypeByRef()) {
						assert(cmdInfo.tmpVars.size() == 1);

						string nullTmp = TAVR_TMP + ToString(cmdInfo.tmpVars[0]);
						output << tabs << nullTmp << " = " << val << ";" << endl;
						output << tabs << TAVR_ASSERT_NULL_PTR << "(" << nullTmp << ");" << endl;

						swap(val, nullTmp);
						val += ".ptr";
						val += "->";
					} else {
						val += ".";
					}

					val += cmd.info.name.str();
					if (cmd.isUsed) {
						swap(info[i].value, val);
					} else {
						output << tabs << val << ";" << endl;
					}
				}
				break;

			case EVAL_COMMAND::TYPE_PRE_INC:
				assert(cmd.args.size() == 1);
				if (cmd.isUsed) {
					std::string val = "++";
					val += info[cmd.args[0]].value;
					swap(cmdInfo.value, val);
				} else {
					output << tabs << "++" << info[cmd.args[0]].value << ";" << endl;
				}
				break;

			case EVAL_COMMAND::TYPE_PRE_DEC:
				assert(cmd.args.size() == 1);
				if (cmd.isUsed) {
					std::string val = "--";
					val += info[cmd.args[0]].value;
					swap(cmdInfo.value, val);
				} else {
					output << tabs << "--" << info[cmd.args[0]].value << ";" << endl;
				}
				break;

			case EVAL_COMMAND::TYPE_POST_INC:
				assert(cmd.args.size() == 1);
				if (cmd.isUsed) {
					std::string val = info[cmd.args[0]].value;
					val += "++";
					swap(cmdInfo.value, val);
				} else {
					output << tabs << info[cmd.args[0]].value << "++" << ";" << endl;
				}
				break;

			case EVAL_COMMAND::TYPE_POST_DEC:
				assert(cmd.args.size() == 1);
				if (cmd.isUsed) {
					std::string val = info[cmd.args[0]].value;
					val += "--";
					swap(cmdInfo.value, val);
				} else {
					output << tabs << info[cmd.args[0]].value << "--" << ";" << endl;
				}
				break;

			case EVAL_COMMAND::TYPE_TYPE_CAST:
				assert(cmd.args.size() == 1);

				if (cmd.isUsed) {
					cmdInfo.value = "(";
					cmdInfo.value += GetTypeName(cmd.typeinfo);
					cmdInfo.value += ")";
					cmdInfo.value += info[cmd.args[0]].value;
				}
				break;

			case EVAL_COMMAND::TYPE_ASSIGN:
				{
					assert(cmd.args.size() == 2);

#ifdef _DEBUG
					output << tabs << "// operator=(" << project->Type2Str(code[cmd.args[0]].typeinfo)
						<< ", " << project->Type2Str(code[cmd.args[1]].typeinfo) << ")" << endl;
#endif

					assert(cmd.info.proc != NULL || cmd.typeinfo.IsTypeByRef() == code[cmd.args[0]].typeinfo.IsTypeByRef());
					assert(cmd.info.proc != NULL || cmd.typeinfo.IsTypeByRef() == code[cmd.args[1]].typeinfo.IsTypeByRef());

					const TYPEINFO & rightType = code[cmd.args[1]].typeinfo;

					bool needResTmp = (cmd.isUsed || cmd.typeinfo.IsTypeByRef()) && !rightType.IsNull();

					assert((int)cmdInfo.tmpVars.size() >= (int)needResTmp);

					string & argValue = info[cmd.args[0]].value;

					string left;
					swap(left, argValue);
					if (needResTmp) {
						swap(left, argValue);
						left = TAVR_TMP + ToString(cmdInfo.tmpVars.back());
						output << tabs << left << " = &" << argValue << ";" << endl;
						left = "(*" + left + ")";
					}

					// check right type for 'null'
					if (rightType.IsNull()) {
						PrintDestructor(cmd.typeinfo, left);
						PrintVarDefInit(cmd.typeinfo, left);
					} else {
						PrintAssignOp(true, cmd.typeinfo, left, cmdInfo, localInfo, &cmd, &code);
					}

					if (!needResTmp) {
						swap(left, argValue);
					} else {
						cmdInfo.value = left;
					}
				}
				break;

			case EVAL_COMMAND::TYPE_SEND:
				assert(cmd.args.size() == 2);
				assert(!cmd.isUsed);
				assert(cmd.info.proc != NULL);
#ifdef _DEBUG
				output << tabs << "// send " << project->Type2Str(code[cmd.args[0]].typeinfo)
					<< " to " << project->Type2Str(code[cmd.args[1]].typeinfo) << endl;
#endif
				output << tabs << TAVR_ADD_POST_LAYER "();" << endl;
				output << tabs << TAVR_CUR_EVENT " = *(" TAVR_NULL_TYPE "*)&" << info[cmd.args[0]].value << ";" << endl;
				output << tabs << TAVR_INC_REF "(" TAVR_CUR_EVENT ");" << endl;
				output << tabs << GetProcName(cmd.info.proc) << "(&" << info[cmd.args[1]].value << ");" << endl;
				output << tabs << TAVR_DEC_REF "(" TAVR_CUR_EVENT ");" << endl;
				output << tabs << TAVR_CUR_EVENT ".ptr = " TAVR_CUR_EVENT ".ref = 0;" << endl;
				output << tabs << TAVR_REMOVE_POST_LAYER "();" << endl;
				break;

			case EVAL_COMMAND::TYPE_POST:
				assert(cmd.args.size() == 2);
				assert(!cmd.isUsed);
				assert(cmd.info.proc != NULL);
#ifdef _DEBUG
				output << tabs << "// post " << project->Type2Str(code[cmd.args[0]].typeinfo)
					<< " to " << project->Type2Str(code[cmd.args[1]].typeinfo) << endl;
#endif
				output << tabs << TAVR_INC_REF "(" << info[cmd.args[0]].value << ");" << endl;
				output << tabs << TAVR_INC_REF "(" << info[cmd.args[1]].value << ");" << endl;
				output << tabs << TAVR_POST_EVENT "(*(" TAVR_NULL_TYPE "*)&"
					<< info[cmd.args[0]].value << ", *(" TAVR_NULL_TYPE "*)&"
					<< info[cmd.args[1]].value << ", (" TAVR_AUTO_PROC ")"
					<< GetProcName(cmd.info.proc) << ");" << endl;
				break;

			case EVAL_COMMAND::TYPE_ON:
				assert(cmd.typeinfo.IsEvent() || cmd.typeinfo.IsNull());
				assert(cmd.typeinfo.IsNull() || cmd.typeinfo.info.type->eventId >= 0);
				if (!cmd.typeinfo.IsNull()) {
#ifdef _DEBUG
					output << tabs << "// on " << project->Type2Str(cmd.typeinfo) << endl;
#endif
					output << tabs << "if (" TAVR_CUR_EVENT ".ref != 0 && ((int*)"
						TAVR_CUR_EVENT ".ref[" TAVR_TYPEINFO_OFFSET "])[" TAVR_EVENT_ID_OFFSET "] == "
						<< cmd.typeinfo.info.type->eventId << ")" << endl;
				} else {
#ifdef _DEBUG
					output << tabs << "// otherwise" << endl;
#endif
					output << tabs << "if (" TAVR_CUR_EVENT ".ref != 0)" << endl;
				}
				break;

			case EVAL_COMMAND::TYPE_OUT_CALL:
				{
					assert(cmd.args.size() == 1);
					assert(cmd.typeinfo.IsEvent());
#ifdef _DEBUG
					output << tabs << "// out " << project->Type2Str(cmd.typeinfo) << endl;
#endif
					assert(localInfo.curProc->owner.type != NULL);
					TYPE * curType = localInfo.curProc->owner.type;
					const char * thisVal = curType->IsTypeByValue()
						? TAVR_THIS "->" : TAVR_THIS "->ptr->";

					COMMAND_INFO & eventInfo = info[cmd.args[0]];

					output << tabs << TAVR_INC_REF "(" << eventInfo.value << ");" << endl;

					output << tabs << TAVR_OUT_CALL "(" << thisVal
						<< TAVR_OUT_PREFIX << cmd.info.name.str() << ", *(" TAVR_NULL_TYPE "*)&"
						<< eventInfo.value << ");" << endl;
				}
				break;

			case EVAL_COMMAND::TYPE_REG_POST:
			case EVAL_COMMAND::TYPE_REG_SEND:
			case EVAL_COMMAND::TYPE_UNREG:
				{
					assert(cmd.args.size() == 2);
					assert(cmd.info.proc != NULL);
					assert(cmd.info.proc->type == METHOD::TYPE_AUTO_PROC);

					const bool isPost = (cmd.type == EVAL_COMMAND::TYPE_REG_POST);
					const bool isUnreg = (cmd.type == EVAL_COMMAND::TYPE_UNREG);

#ifdef _DEBUG
					output << tabs << "// ";
					output << project->Type2Str(code[cmd.args[0]].typeinfo);
					if (isUnreg) {
						output << " -= ";
					} else if (isPost) {
						output << " += post ";
					} else {
						output << " += send ";
					}
					output << project->Type2Str(code[cmd.args[1]].typeinfo) << endl;
#endif

					const char * subVal = code[cmd.args[0]].typeinfo.IsTypeByRef()
						? ".ptr->" : ".";

					const EVAL_COMMAND & autoCmd = code[cmd.args[1]];
					const COMMAND_INFO & autoInfo = info[cmd.args[1]];

					output << tabs << TAVR_INC_REF "(" << autoInfo.value << ");" << endl;

					if (!isUnreg) {
						output << tabs << TAVR_ADD_OUT_LIST "(&";
					} else {
						output << tabs << TAVR_REMOVE_OUT_LIST "(&";
					}
					
					assert(autoCmd.typeinfo.IsAuto());

					output << info[cmd.args[0]].value << subVal << TAVR_OUT_PREFIX 
						<< cmd.info.name.str() << ", *(" TAVR_NULL_TYPE "*)&"
						<< autoInfo.value;
					
					if (!isUnreg) {
						output << ", (" TAVR_AUTO_PROC ")"
							<< GetProcName(cmd.info.proc)
							<< ", " << (int)isPost;
					}
					output << ");" << endl;
				}
				break;

			case EVAL_COMMAND::TYPE_EXPR_IF_COND:
				assert(cmdInfo.tmpVars.size() == 1);
				assert(cmd.args.size() == 1);
				output << tabs << "if (" << info[cmd.args[0]].value << ") {" << endl;
				AddTab();
				break;

			case EVAL_COMMAND::TYPE_EXPR_IF_ELSE:
				{
					assert(cmd.args.size() == 2);
					COMMAND_INFO & condInfo = info[cmd.args[0]];
					assert(condInfo.tmpVars.size() == 1);
					output << tabs << TAVR_TMP << condInfo.tmpVars[0] << " = " << info[cmd.args[1]].value << ";" << endl;

					assert(localInfo.forceReleaseTmpVars.find(i) != localInfo.forceReleaseTmpVars.end());
					PENDING_TMP_VAR_ARRAY & arr = localInfo.forceReleaseTmpVars[i];
					FinalizeTmpVars(arr, localInfo);

					RemoveTab();
					output << tabs << "} else {" << endl;
					AddTab();
				}
				break;
	
			case EVAL_COMMAND::TYPE_EXPR_IF_RES:
				{
					assert(cmd.args.size() == 2);
					const EVAL_COMMAND & elseCmd = code[cmd.args[0]];
					assert(elseCmd.type == EVAL_COMMAND::TYPE_EXPR_IF_ELSE);
					COMMAND_INFO & condInfo = info[elseCmd.args[0]];
					assert(condInfo.tmpVars.size() == 1);
					cmdInfo.value = TAVR_TMP + ToString(condInfo.tmpVars[0]);

					output << tabs << cmdInfo.value << " = " << info[cmd.args[1]].value << ";" << endl;

					assert(localInfo.forceReleaseTmpVars.find(i) != localInfo.forceReleaseTmpVars.end());
					PENDING_TMP_VAR_ARRAY & arr = localInfo.forceReleaseTmpVars[i];
					FinalizeTmpVars(arr, localInfo);

					RemoveTab();
					output << tabs << "}" << endl;
				}
				break;

			case EVAL_COMMAND::TYPE_EQ:
			case EVAL_COMMAND::TYPE_NOT_EQ:
				{
					assert(cmd.args.size() == 2);

					std::string val;
					
					const EVAL_COMMAND & leftCmd = code[cmd.args[0]];
					if (leftCmd.type != EVAL_COMMAND::TYPE_CONSTANT || !leftCmd.typeinfo.IsNull()) {
						val += info[ cmd.args[0] ].value;
						if (leftCmd.typeinfo.IsTypeByRef()) {
							val += ".ref";
						}
					} else {
						val += "0";
					}

					val += cmd.type == EVAL_COMMAND::TYPE_EQ
						? " == " : " != ";

					const EVAL_COMMAND & rightCmd = code[cmd.args[1]];
					if (rightCmd.type != EVAL_COMMAND::TYPE_CONSTANT || !rightCmd.typeinfo.IsNull()) {
						val += info[ cmd.args[1] ].value;
						if (rightCmd.typeinfo.IsTypeByRef()) {
							val += ".ref";
						}
					} else {
						val += "0";
					}

					if (cmd.isUsed) {
						val += ")";
						swap(cmdInfo.value, "(" + val);
					} else {
						output << tabs << val << ";" << endl;
					}
				}
				break;

#define CASE_NATIVE_UNAR(EC_TYPE, VAL) \
			case EVAL_COMMAND::TYPE_##EC_TYPE: \
				{ \
					assert(cmd.args.size() == 1); \
					std::string val = VAL " "; \
					val += info[ cmd.args[0] ].value; \
					\
					if (cmd.isUsed) { \
						val += ")"; \
						swap(cmdInfo.value, "(" + val); \
					} else { \
						output << tabs << val << ";" << endl; \
					} \
				} \
				break

			CASE_NATIVE_UNAR(NEG,		"-");
			CASE_NATIVE_UNAR(LOG_NOT,	"!");
			CASE_NATIVE_UNAR(BIT_NOT,	"~");

#define CASE_NATIVE_BINARY(EC_TYPE, VAL) \
			case EVAL_COMMAND::TYPE_##EC_TYPE: \
				{ \
					assert(cmd.args.size() == 2); \
					std::string val = info[ cmd.args[0] ].value; \
					val += " " VAL " "; \
					val += info[ cmd.args[1] ].value; \
					\
					if (cmd.isUsed) { \
						val += ")"; \
						swap(cmdInfo.value, "(" + val); \
					} else { \
						output << tabs << val << ";" << endl; \
					} \
				} \
				break

			CASE_NATIVE_BINARY(ADD,    		"+");
			CASE_NATIVE_BINARY(SUB,    		"-");
			CASE_NATIVE_BINARY(MUL,    		"*");
			CASE_NATIVE_BINARY(DIV,    		"/");
			CASE_NATIVE_BINARY(MOD,    		"%");

			CASE_NATIVE_BINARY(LOG_OR,		"||");
			CASE_NATIVE_BINARY(LOG_AND,		"&&");
			CASE_NATIVE_BINARY(LESS,		"<");
			CASE_NATIVE_BINARY(LESS_EQ,		"<=");
			CASE_NATIVE_BINARY(GREATER,		">");
			CASE_NATIVE_BINARY(GREATER_EQ,	">=");

			default:
				assert(!"Unknown output command type");
				break;
		}

		// finalize pending tmp vars
		assert(cmdInfo.argSet != -1);
		FinalizeTmpVars(cmdInfo.pendingTmpVars, localInfo);
	}
}

/*************************************************************************/
/* TARGET_C::PreGenerate()                                               */
/*                                                                       */
/*************************************************************************/
bool TARGET_C::PreGenerate(void)
{
	settings = project->settings->GetSettings("project/target_c");
	if (settings == NULL) {
		project->GetError() << "ERROR: section 'target_c' not found" << endl;
		return false;
	}

	moduleList.clear();

	name	= project->settings->GetValue("project/name");

	tmp	= settings->GetValue("temp", "tmp");
	out	= settings->GetValue("output", "out");

	ecg::FolderCreate(tmp.c_str());
	ecg::FolderCreate(out.c_str());

	tavrRuntime = settings->GetValue("tavr_runtime");
	if (tavrRuntime.empty()) {
		project->GetError() << "ERROR: variable 'tavr_runtime' should contain runtime headers path" << endl;
		return false;
	}

	return true;
}

/*************************************************************************/
/* TARGET_C::GenerateTypeinfo()                                          */
/*                                                                       */
/*************************************************************************/
bool TARGET_C::GenerateTypeinfo(std::string & outputName)
{
	outputName = tmp + "\\_typeinfo.c";

	// TODO: support project settings (temp dir and so on)
	output.open(outputName.c_str());

	if (!output) {
		return false;
	}

	output << "#include <tavr_runtime.h>" << endl;

	GLOBAL_INFO globalInfo;
	INFO localInfo;

	for (std::set<TYPEINFO>::iterator iter = typeinfoTypes.begin(); iter != typeinfoTypes.end(); ++iter) {
		TYPEINFO & ti = *iter;

		// ignore all value types
		if (!ti.IsTypeByRef() || ti.IsNull()) {
			continue;
		}

		// ignore static types
		if (ti.array == 0 && ti.info.type->attribs.is(ATTRIB_STATIC)) {
			continue;
		}

		if (ti.array > 0) {
			// generate array destructor
			PreProcessTypePrintDecl(ti, globalInfo);

			PrintProcHeaderArrDestr(ti);
			output << endl << "{" << endl;
			AddTab();
			PrintArrDestrBody(ti, "(*" TAVR_THIS ")");
			RemoveTab();
			output << "}" << endl;
		} else if (ti.info.type->specs[TYPE::SPEC_DESTR] != NULL) {
			// forward declaration for corresponding destructors
			PrintProcHeader(ti.info.type->specs[TYPE::SPEC_DESTR], globalInfo);
			output << ";" << endl;
		}

		const string infoName = TAVR_TYPEINFO + GetTypeName(ti, true);
		output << "void * " << infoName << "[] = {" << endl;
		// type name
		output << "	\"" << project->Type2Str(ti) << "\"," << endl;
		// destructor
		if (ti.array > 0) {
			output << "	" TAVR_ARRAY_DESTR << GetTypeName(ti, true) << "," << endl;
		} else {
			assert(ti.info.type->specs[TYPE::SPEC_DESTR] != NULL);
			output << "	" << GetProcName(ti.info.type->specs[TYPE::SPEC_DESTR]) << "," << endl;

			if (ti.IsEvent()) {
				output << "	(void *)" << ti.info.type->eventId << "," << endl;
			}
		}
		output << "};" << endl;
	}

	output.close();

	return true;
}

/************************************************************************/
/* tavr::TARGET_C::PostGenerate
/*
/************************************************************************/
bool TARGET_C::PostGenerate(void)
{
	std::string compiler = settings->GetValue("compiler");

	std::string compilationType = settings->GetValue("compilationType", "multiple");
	if (compilationType != "multiple" && compilationType != "single") {
		project->GetError() << "ERROR: Unknown compilation type '" << compilationType << "'" << endl;
		return false;
	}

	std::string subSystem = settings->GetValue("subsystem", "console");
#ifdef _DEBUG
	subSystem = "console"; // always create console app in debug (for debug output)
#endif

	std::string outSubSystem;
	if (subSystem == "console") {
		outSubSystem = "/SUBSYSTEM:CONSOLE";
	} else if (subSystem == "windows") {
		outSubSystem = "/SUBSYSTEM:WINDOWS";
	} else {
		project->GetError() << "ERROR: Unknown subsytem '" << subSystem << "'" << endl;
		return false;
	}

	std::string	pre_exec;

	std::string tavrRuntimeOptions;
	std::string options;
	std::string	exec;
	std::string	link_options;

	if (compiler == "vc70" || compiler == "vc71" || compiler == "vc80") {
		if (compiler == "vc70") {
			pre_exec = "\"%VS70COMNTOOLS%vsvars32.bat\"";
		} else if (compiler == "vc71") {
			pre_exec = "\"%VS71COMNTOOLS%vsvars32.bat\"";
		} else if (compiler == "vc80") {
			pre_exec = "\"%VS80COMNTOOLS%vsvars32.bat\"";
		}

		exec = "cl.exe";
		// /MD <-> /MT
#ifndef _DEBUG
		options = "/O2 /GL /D \"WIN32\" /D \"NDEBUG\" /D \"_CONSOLE\" /FD /EHsc /MT /Zp1 /Fo\"";
#else
		options = "/Od /GL /D \"WIN32\" /D \"_DEBUG\" /D \"_CONSOLE\" /FD /EHsc /MTd /Zi /Zp1 /Fo\"";
#endif
		options += tmp + "\\\\\" /Fd\"" + tmp + "\\vc.pdb\" /W3 /nologo /Wp64 /Zi /TC";

#ifndef _DEBUG
		link_options = "/link /OUT:\"" + out + "\\" + name + ".exe\" /PDB:\"" + out + "\\" + name + ".pdb\" " + outSubSystem + " /MACHINE:X86";
#else
		link_options = "/link /OUT:\"" + out + "\\" + name + ".exe\" /PDB:\"" + out + "\\" + name + ".pdb\" " + outSubSystem + " /MACHINE:X86 /DEBUG";
#endif

		// add additional lib path
		for (int i=0; i<(int)settings->subSettings.size(); ++i) {
			PROJECT_SETTINGS * s = settings->subSettings[i];
			if (s->name != "lib_path") {
				continue;
			}

			link_options += " /LIBPATH:\"" + s->value + "\"";
		}

		tavrRuntimeOptions = "/I \"" + tavrRuntime + "\"";
	} else if (!compiler.empty()) {
		project->GetError() << "ERROR: Unknown compiler type '" << compiler << "'" << endl;
		return false;
	}

	//////////////////////////////////////////////////////////////////////////
	// COMPILE .C FILES

	pre_exec				= settings->GetValue("pre_exec",	pre_exec);
	exec					= settings->GetValue("exec",		exec);
	options					= settings->GetValue("options",	options);

	std::string batchName	= settings->GetValue("batch_name", "build.bat");
	string batchPath = tmp + "\\" + batchName;
	ofstream batch(batchPath.c_str());
	if (!batch) {
		project->GetError() << "ERROR: Failed to create batch file \"" << batchPath << "\"" << endl;
		return false;
	}

	batch << "@echo off" << endl;
	if (!pre_exec.empty()) {
		batch << "call " << pre_exec << " > nul" << endl;
	}
	batch << exec << " " << options;

	batch << " " << tavrRuntimeOptions;

	for (int i=0; i<(int)moduleList.size(); ++i) {
		batch << " " << tmp << "\\" << moduleList[i]->name << ".c";
	}
	batch << " " << tmp << "\\_typeinfo.c";

	// linking options
	batch << " " << link_options;

	for (int i=0; i<(int)settings->subSettings.size(); ++i) {
		if (settings->subSettings[i]->name == "lib") {
			batch << " \"" << settings->subSettings[i]->value << "\"";
		}
	}

	batch << endl;
	batch.close();

	int res = system(("\"" + batchPath + "\"").c_str());

	if (res != 0) {
		project->GetError() << "ERROR: Compilation failed" << endl;
		return false;
	}

	return true;
}

} // namespace tavr
